#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Manpreet Singh <manpreet.singh.dev@proton.me>
# SPDX-FileCopyrightText: 2024 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

from abc import abstractmethod
import abc
import cmath
from enum import Enum
from typing import Dict, List, Optional, Tuple, Type

import math
import inkex
from inkex.paths import Line, Move, ZoneClose, Curve, Arc, move
from inkex import Path, Vector2d, BoundingBox

from ..parser.types import AFObject
from .util import bbox_helper, safe_div, interpolate_float

# TODO: For shapes, the stroke's start point should be at the top right corner of the shape


class AFShape:
    _id: int = 0

    def __init__(self, b_box: BoundingBox):
        self.b_box = b_box

        self.id = self.__class__._id
        self.__class__._id += 1

    @classmethod
    def from_af(cls, shpe: AFObject, b_box: BoundingBox) -> AFShape:
        stype = shpe.get_type()
        assert stype is not None, "Undefined Shape type"
        return cls(b_box)


class AFPointsShapeBase(AFShape):
    def __init__(self, b_box: BoundingBox):
        super().__init__(b_box)

    @property
    @abstractmethod
    def points(self) -> List[Vector2d]: ...


class AFEllipseBase(AFShape):
    def __init__(self, b_box: BoundingBox):
        super().__init__(b_box)

    @property
    def radii(self) -> Vector2d:
        b_box = self.b_box
        return Vector2d(b_box.width / 2, b_box.height / 2)


class AFCornerPos(Enum):
    TL = 0
    TR = 1
    BR = 2
    BL = 3


class AFCornerType(Enum):
    ROUNDED = 0
    STRAIGHT = 1
    CONCAVE = 2
    CUTOUT = 3
    NONE = 4


corner_map = {
    AFCornerType.NONE: "F",  # radius will be set to 0
    AFCornerType.ROUNDED: "F",
    AFCornerType.CONCAVE: "IF",
    AFCornerType.STRAIGHT: "C",
}


class AFRectangle(AFShape):
    def __init__(
        self,
        b_box: BoundingBox,
        radii: List[float] = [0.25] * 4,
        corner_types: List[AFCornerType] = [AFCornerType.NONE] * 4,
        lock: bool = True,
        absolute_size: bool = False,
    ):
        super().__init__(b_box)
        self.lock = lock
        self.absolute_size = absolute_size
        self.corner_types = corner_types
        self.radii = radii

    @classmethod
    def from_af(cls, shpe: AFObject, b_box: BoundingBox) -> AFRectangle:
        sup = super().from_af(shpe, b_box)
        assert shpe.get_type() == "ShNR", "Not a rectangle"
        assert isinstance(sup, AFRectangle)

        if "Lock" in shpe:
            sup.lock = shpe["Lock"]
        if "AbSz" in shpe:
            sup.absolute_size = shpe["AbSz"]
        if "CTyp" in shpe:
            sup.corner_types = [AFCornerType(ct.id) for ct in shpe["CTyp"]]
        if "ShCR" in shpe:
            sup.radii = shpe["ShCR"]
        return sup

    @property
    def radii(self) -> List[float]:
        radii = self._radii
        if not self.absolute_size:
            radii = [r * min(self.b_box.width, self.b_box.height) for r in radii]
        if self.lock:
            radii = [radii[0]] * 4

        radii = [
            r if c != AFCornerType.NONE else 0.0
            for r, c in zip(radii, self.corner_types)
        ]

        if (
            radii[0] + radii[1] > self.b_box.width
            or radii[2] + radii[3] > self.b_box.width
            or radii[3] + radii[0] > self.b_box.height
            or radii[1] + radii[2] > self.b_box.height
        ):
            ratio = min(
                safe_div(self.b_box.width, (radii[0] + radii[1]), float("inf")),
                safe_div(self.b_box.width, (radii[2] + radii[3]), float("inf")),
                safe_div(self.b_box.height, (radii[3] + radii[0]), float("inf")),
                safe_div(self.b_box.height, (radii[1] + radii[2]), float("inf")),
            )
            if ratio != float("inf"):
                radii = [r * ratio for r in radii]
        return radii

    @radii.setter
    def radii(self, value: List[float]):
        assert len(value) == 4
        self._radii = value

    @property
    def corner_types(self) -> List[AFCornerType]:
        if self.lock:
            return [AFCornerType(self._corner_types[0])] * 4
        return [ct for ct in self._corner_types]

    @corner_types.setter
    def corner_types(self, value: List[AFCornerType]):
        assert len(value) == 4 and all(isinstance(ct, AFCornerType) for ct in value)
        self._corner_types = value

    def _corner_path(self, pos: AFCornerPos) -> List[Tuple[str, List[float]]]:
        corner_type = self.corner_types[pos.value]
        radius = self.radii[pos.value]

        if radius <= 0:
            return []

        dx = radius if pos in [AFCornerPos.TL, AFCornerPos.TR] else -radius
        dy = radius if pos in [AFCornerPos.TR, AFCornerPos.BR] else -radius

        if corner_type == AFCornerType.STRAIGHT:
            return [("l", [dx, dy])]

        if corner_type in [AFCornerType.ROUNDED, AFCornerType.CONCAVE]:
            # arc: a rx ry x-axis-rotation large-arc-flag sweep-flag dx dy
            is_round = corner_type == AFCornerType.ROUNDED
            return [("a", [radius, radius, 0, 0, int(is_round), dx, dy])]

        if corner_type == AFCornerType.CUTOUT:
            return (
                [("h", [dx]), ("v", [dy])]
                if pos in [AFCornerPos.TL, AFCornerPos.BR]
                else [("v", [dy]), ("h", [dx])]
            )

        return (
            [("h", [dx]), ("v", [dy])]
            if pos not in [AFCornerPos.TL, AFCornerPos.BR]
            else [("v", [dy]), ("h", [dx])]
        )

    def _corner_path_effect(self) -> inkex.PathEffect:
        assert all(ct in corner_map for ct in self.corner_types), (
            f"Invalid corner types {[ct for ct in self.corner_types if ct not in corner_map]}"
        )

        params = " @ ".join(
            f"{corner_map[c]},1,0,1,0,{r / length},0,1"
            for c, r, length in zip(
                self.corner_types, self.radii, [self.b_box.width, self.b_box.height] * 2
            )
        )
        return inkex.PathEffect.new(
            effect="fillet_chamfer",
            lpeversion="1",
            method="arc",
            flexible="true",  # radius in %
            satellites_param=params,  # Inkscape 1.2
            nodesatellites_param=params,  # Inkscape 1.3
        )


class AFEllipse(AFEllipseBase):
    @classmethod
    def from_af(cls, shpe: AFObject, b_box: BoundingBox) -> AFEllipse:
        sup = super().from_af(shpe, b_box)
        assert shpe.get_type() == "ShpE", "Not an ellipse"
        assert isinstance(sup, AFEllipse)
        return sup


class AFTriangle(AFPointsShapeBase):
    def __init__(self, b_box: BoundingBox, pos: float = 0.5):
        super().__init__(b_box)
        self.pos = pos

    @classmethod
    def from_af(cls, shpe: AFObject, b_box: BoundingBox) -> AFTriangle:
        sup = super().from_af(shpe, b_box)
        assert shpe.get_type() == "ShpT", "Not a triangle"
        assert isinstance(sup, AFTriangle)

        if "Pos " in shpe:
            sup.pos = shpe["Pos "]

        return sup

    @property
    def points(self) -> List[Vector2d]:
        b_box = self.b_box
        return [
            Vector2d(b_box.left + (b_box.width * self.pos), b_box.top),
            Vector2d(b_box.right, b_box.bottom),
            Vector2d(b_box.left, b_box.bottom),
        ]


class AFDiamond(AFPointsShapeBase):
    def __init__(self, b_box: BoundingBox, pos: float = 0.5):
        super().__init__(b_box)
        self.pos = pos

    @classmethod
    def from_af(cls, shpe: AFObject, b_box: BoundingBox) -> AFDiamond:
        sup = super().from_af(shpe, b_box)
        assert shpe.get_type() == "ShpD", "Not a diamond"
        assert isinstance(sup, AFDiamond)

        if "Pos " in shpe:
            sup.pos = shpe["Pos "]

        return sup

    @property
    def points(self) -> List[Vector2d]:
        b_box = self.b_box
        hpos, vpos = 0.5, 1 - self.pos
        return [
            Vector2d(b_box.left + (b_box.width * hpos), b_box.top),
            Vector2d(b_box.right, b_box.top + (b_box.height * vpos)),
            Vector2d(b_box.left + (b_box.width * hpos), b_box.bottom),
            Vector2d(b_box.left, b_box.top + (b_box.height * vpos)),
        ]


class AFTrapezoid(AFPointsShapeBase):
    def __init__(self, b_box: BoundingBox, posl: float = 0.25, posr: float = 0.75):
        super().__init__(b_box)
        self.posl = posl
        self.posr = posr

    @classmethod
    def from_af(cls, shpe: AFObject, b_box: BoundingBox) -> AFTrapezoid:
        sup = super().from_af(shpe, b_box)
        assert shpe.get_type() == "ShTz", "Not a trapezoid"
        assert isinstance(sup, AFTrapezoid)

        if "PosL" in shpe:
            sup.posl = shpe["PosL"]
        if "PosR" in shpe:
            sup.posr = shpe["PosR"]

        return sup

    @property
    def points(self) -> List[Vector2d]:
        b_box = self.b_box
        posl, posr = self.posl, self.posr
        return [
            Vector2d(b_box.right, b_box.bottom),
            Vector2d(b_box.left, b_box.bottom),
            Vector2d(b_box.left + (b_box.width * posl), b_box.top),
            Vector2d(b_box.left + (b_box.width * posr), b_box.top),
        ]


class AFPie(AFEllipseBase):
    def __init__(
        self,
        b_box: BoundingBox,
        start_angle: float = math.pi / 2,
        end_angle: float = 0.0,
        inner_radius: float = 0.0,
    ):
        super().__init__(b_box)
        self.start_angle = start_angle
        self.end_angle = end_angle
        self.inner_radius = inner_radius

    @classmethod
    def from_af(cls, shpe: AFObject, b_box: BoundingBox) -> AFPie:
        sup = super().from_af(shpe, b_box)
        assert shpe.get_type() == "ShPi", "Not a pie"
        assert isinstance(sup, AFPie)

        if "AngS" in shpe:
            sup.start_angle = shpe["AngS"]
        if "AngE" in shpe:
            sup.end_angle = shpe["AngE"]
        if "IRad" in shpe:
            sup.inner_radius = shpe["IRad"]

        return sup


class AFSquareStar(AFPointsShapeBase, AFEllipseBase):
    def __init__(
        self,
        b_box: BoundingBox,
        sides: int = 5,
        cutout: float = 0.5,
    ):
        super().__init__(b_box)
        self.sides = sides
        self.cutout = cutout

    @classmethod
    def from_af(cls, shpe: AFObject, b_box: BoundingBox) -> AFSquareStar:
        sup = super().from_af(shpe, b_box)
        assert shpe.get_type() == "ShSS", "Not a square star"
        assert isinstance(sup, AFSquareStar)

        if "COut" in shpe:
            sup.cutout = shpe["COut"]
        if "Side" in shpe:
            sup.sides = shpe["Side"]

        return sup

    @property
    def points(self) -> List[Vector2d]:
        center, radii, sides, cutout = (
            self.b_box.center,
            self.radii,
            self.sides,
            self.cutout,
        )
        ellipse_points = points_on_ellipse(radii, center, sides)

        points: List[Vector2d] = []
        for p1, p2 in zip(ellipse_points, ellipse_points[1:] + [ellipse_points[0]]):
            mid = (p1[0] + p2[0]) / 2, (p1[1] + p2[1]) / 2

            points.append(
                Vector2d(
                    interpolate_float(p1[0], mid[0], cutout),
                    interpolate_float(p1[1], mid[1], cutout),
                )
            )
            if cutout != 0:
                points.extend(
                    (
                        Vector2d(
                            interpolate_float(p2[0], mid[0], cutout),
                            interpolate_float(p2[1], mid[1], cutout),
                        ),
                        Vector2d(
                            interpolate_float(p2[0], center[0], cutout),
                            interpolate_float(p2[1], center[1], cutout),
                        ),
                    )
                )

        return points


class AFDoubleStar(AFPointsShapeBase, AFEllipseBase):
    def __init__(
        self,
        b_box: BoundingBox,
        points: int = 5,
        inner_radius: float = 0.525731086730957,  # default in Designer
        point_radius: float = 0.80901700258255,  # default in Designer
    ):
        super().__init__(b_box)
        self.num_points = points
        self.inner_radius = inner_radius
        self.point_radius = point_radius

    @classmethod
    def from_af(cls, shpe: AFObject, b_box: BoundingBox) -> AFDoubleStar:
        sup = super().from_af(shpe, b_box)
        assert shpe.get_type() == "ShDS", "Not a square star"
        assert isinstance(sup, AFDoubleStar)

        if "Pnts" in shpe:
            sup.num_points = shpe["Pnts"]
        if "IRad" in shpe:
            sup.inner_radius = shpe["IRad"]
        if "PRad" in shpe:
            sup.point_radius = shpe["PRad"]

        return sup

    @property
    def points(self) -> List[Vector2d]:
        center, radii, npoints = self.b_box.center, self.radii, self.num_points
        dt = 2 * math.pi / npoints
        outer_p1 = points_on_ellipse(radii, center, npoints)
        outer_p2 = points_on_ellipse(radii * self.point_radius, center, npoints, dt / 2)
        inner_p = points_on_ellipse(
            radii * self.inner_radius, center, npoints * 2, dt / 4
        )

        return [
            p
            for pts in zip(outer_p1, inner_p[::2], outer_p2, inner_p[1::2])
            for p in pts
        ]


def points_on_ellipse(
    radii: Vector2d,
    center: Vector2d,
    count: int,
    offset: float = 0,
) -> List[Vector2d]:
    """Returns a list of `count` number of points on an ellipse."""
    step = 2 * math.pi / count
    return [
        Vector2d(
            radii.x * math.cos(i * step - 90 * math.pi / 180 + offset) + center.x,
            radii.y * math.sin(i * step - 90 * math.pi / 180 + offset) + center.y,
        )
        for i in range(count)
    ]


class AFCurveShape(AFEllipseBase, abc.ABC):
    def get_svg_path(self) -> Path:
        path = self._get_svg_path()
        return path.transform(
            inkex.Transform(scale=tuple(self.radii), translate=tuple(self.b_box.center))
        )

    @abc.abstractmethod
    def _get_svg_path(self): ...


class AFStar(AFCurveShape):
    def __init__(
        self,
        b_box: BoundingBox,
        points: int = 5,
        inner_radius: float = 1,  # maximum if missing
        outer_circle: float = 0.0,
        inner_circle: float = 0.0,
        left_curve: float = 0.0,
        right_curve: float = 0.0,
        curved_edges: bool = False,
    ):
        super().__init__(b_box)
        self.num_points = points
        self.inner_radius = inner_radius
        self.outer_circle = outer_circle
        self.inner_circle = inner_circle
        self.left_curve = left_curve
        self.right_curve = right_curve
        self.curved_edges = curved_edges

    @property
    def inner_radius_clipped(self):
        """point distance from center, limited such that the outer tip is acute
        https://en.wikipedia.org/wiki/Apothem#Finding_the_apothem"""
        return min(self.inner_radius, math.cos(math.pi / self.num_points))

    @classmethod
    def from_af(cls, shpe: AFObject, b_box: BoundingBox) -> AFStar:
        sup = super().from_af(shpe, b_box)
        assert shpe.get_type() == "ShSt", "Not a star"
        assert isinstance(sup, AFStar)

        if "CrcI" in shpe:
            sup.inner_circle = shpe["CrcI"]
        if "CrcO" in shpe:
            sup.outer_circle = shpe["CrcO"]
        if "CrvL" in shpe:
            sup.left_curve = shpe["CrvL"]
        if "CrvR" in shpe:
            sup.right_curve = shpe["CrvR"]
        if "Lgcy" in shpe:
            sup.curved_edges = shpe["Lgcy"]
        if "Pnts" in shpe:
            sup.num_points = shpe["Pnts"]
        if "IRad" in shpe:
            sup.inner_radius = shpe["IRad"]
        return sup

    def _get_rounded_corner_svg_path(self) -> Path:
        # The outer radius is normalized by the maximum possible outer radius (where two adjacent outer circles touch)
        max_outer_circle = math.sin(math.pi / self.num_points) / (
            1 + math.sin(math.pi / self.num_points)
        )

        # The actual outer circle radius
        outer_circle_radius = self.outer_circle * max_outer_circle

        # center points of inner and outer circles
        inner_centers = points_on_ellipse(
            Vector2d(self.inner_radius_clipped, self.inner_radius_clipped),
            Vector2d(0, 0),
            self.num_points,
            -math.pi / self.num_points,
        )
        outer_centers = points_on_ellipse(
            Vector2d(1 - outer_circle_radius, 1 - outer_circle_radius),
            Vector2d(0, 0),
            self.num_points,
        )

        # The inner circle is given as a ratio, but not sure which one. A good fit seems to be
        rin_factor = 1 / (self.num_points + 4) * 3.33

        # However it is capped by two quantities:
        # - the sum of both radii cannot be larger than the distance between outer and inner center
        # - the inner circles can at most touch each other
        inner_circle_radius = min(
            self.inner_circle * rin_factor,
            (outer_centers[0] - inner_centers[0]).length - outer_circle_radius,
            math.sin(math.pi / self.num_points) * self.inner_radius_clipped,
        )

        path = Path()
        outer_centers.append(outer_centers[0])
        inner_centers.append(inner_centers[0])

        for i, (pi, po, pi2) in enumerate(
            zip(inner_centers, outer_centers, inner_centers[1:] + inner_centers[0:1])
        ):
            tp1 = get_internal_tangent_points(
                outer_circle_radius, inner_circle_radius, po, pi
            )
            tp2 = get_internal_tangent_points(
                outer_circle_radius, inner_circle_radius, po, pi2
            )

            if i == 0:
                path.append(Move(*tp1[2]))
            else:
                if not math.isclose(inner_circle_radius, 0):
                    path.append(
                        Arc(
                            inner_circle_radius,
                            inner_circle_radius,
                            0,
                            False,
                            False,
                            *tp1[3],
                        )
                    )
                path.append(Line(*tp1[2]))

            if i < self.num_points:
                if not math.isclose(outer_circle_radius, 0):
                    # Large arc if angle between (tp1[2] - po) and (tp2[0]- po) > 180°
                    # Formula: https://stackoverflow.com/a/16544330
                    v1 = tp1[2] - po
                    v2 = tp2[0] - po
                    dot = v1.dot(v2)
                    det = v1.x * v2.y - v1.y * v2.x

                    path.append(
                        Arc(
                            outer_circle_radius,
                            outer_circle_radius,
                            0,
                            False if math.atan2(det, dot) > 0 else True,
                            True,
                            *tp2[0],
                        )
                    )
                path.append(Line(*tp2[1]))
        path.append(ZoneClose())

        return path

    def _get_path_with_curved_edges(self) -> Path:
        # center points of inner and outer circles
        inner_pts = points_on_ellipse(
            Vector2d(self.inner_radius_clipped, self.inner_radius_clipped),
            Vector2d(0, 0),
            self.num_points,
            math.pi / self.num_points,
        )
        outer_pts = points_on_ellipse(
            Vector2d(1, 1),
            Vector2d(0, 0),
            self.num_points,
        )

        n = self.num_points
        ri_ratio = self.inner_radius_clipped

        def get_edge_handles(right=True) -> Tuple[Optional[complex], Optional[complex]]:
            if right:
                percent = self.right_curve
            else:
                percent = self.left_curve
            if percent == 0:
                return None, None

            # There's probably a geometrical justification to these formulas, but I've figured
            # them out through regression

            v1l = (
                math.sin(math.pi / n) * 2 ** (-1 / 3) * percent
                if percent > 0
                else (-(1 / ri_ratio - 1 / math.cos(math.pi / n)) * ri_ratio * percent)
            )

            v1a = math.pi / 2 if percent < 0 else 0

            v2l = (
                (math.cos(math.pi / n) * (1 - ri_ratio) * ((percent - 0.5) / 0.5) ** 3)
                * 2 ** (-1 / 3)
                if percent > 0.5
                else (
                    -math.tan(math.pi / n) * ri_ratio * percent * (2 + percent)
                    if percent < 0
                    else 0
                )
            )
            v2a = (
                -math.pi / 2 + math.pi / n
                if percent >= 0.5
                else (-math.pi + math.pi / n if percent < 0 else 0)
            )

            if right:
                return cmath.rect(v1l, v1a), cmath.rect(v2l, v2a)
            else:
                return -cmath.rect(v1l, v1a).conjugate(), -cmath.rect(
                    v2l, v2a
                ).conjugate()

        lh1, lh2 = get_edge_handles(False)
        rh1, rh2 = get_edge_handles(True)

        path = Path([Move(outer_pts[0])])

        for i, (po, pi, po2) in enumerate(
            zip(outer_pts, inner_pts, outer_pts[1:] + outer_pts[0:1])
        ):
            angle_incr = cmath.rect(1, 2 * math.pi / n * i)
            if rh1 is not None and rh2 is not None:
                path.append(Curve(po + rh1 * angle_incr, pi + rh2 * angle_incr, pi))
            else:
                # Just draw a line
                path.append(Line(pi))
            angle_incr = cmath.rect(1, 2 * math.pi / n * (i + 1))
            if lh1 is not None and lh2 is not None:
                path.append(Curve(pi + lh2 * angle_incr, po2 + lh1 * angle_incr, po2))
            else:
                # Just draw a line
                path.append(Line(po2))

        path.append(ZoneClose())
        return path

    def _get_svg_path(self) -> Path:
        if not self.curved_edges:
            # Rounded Corners
            return self._get_rounded_corner_svg_path()
        return self._get_path_with_curved_edges()


# https://en.wikipedia.org/wiki/Tangent_lines_to_circles#Tangent_lines_to_two_circles
# https://math.stackexchange.com/a/4048944
def get_internal_tangent_points(
    r1: float, r2: float, c1: Vector2d, c2: Vector2d
) -> List[Vector2d]:
    hypotenuse = (c1 - c2).length
    opp = r1 + r2

    if opp > hypotenuse:
        if not math.isclose(opp, hypotenuse):
            raise ValueError(
                f"math domain error. circles intersect ({r1} + {r2} = {opp}) > {hypotenuse}"
            )
        opp = hypotenuse
    theta = (
        math.atan2(c2.y - c1.y, c2.x - c1.x) + math.asin(opp / hypotenuse) - math.pi / 2
    )
    theta2 = (
        math.atan2(c2.y - c1.y, c2.x - c1.x) - math.asin(opp / hypotenuse) + math.pi / 2
    )

    return [
        Vector2d(c1.x + r1 * math.cos(theta), c1.y + r1 * math.sin(theta)),
        Vector2d(
            c2.x + r2 * math.cos(theta + math.pi),
            c2.y + r2 * math.sin(theta + math.pi),
        ),
        Vector2d(c1.x + r1 * math.cos(theta2), c1.y + r1 * math.sin(theta2)),
        Vector2d(
            c2.x + r2 * math.cos(theta2 + math.pi),
            c2.y + r2 * math.sin(theta2 + math.pi),
        ),
    ]


class AFCrescent(AFCurveShape):
    def __init__(
        self,
        b_box: BoundingBox,
        curve_l: float = -100,
        curve_r: float = 100,
    ):
        super().__init__(b_box)
        self.curve_l = curve_l
        self.curve_r = curve_r

    @classmethod
    def from_af(cls, shpe: AFObject, b_box: BoundingBox) -> AFCrescent:
        sup = super().from_af(shpe, b_box)
        assert shpe.get_type() == "ShCr", "Not a crescent"
        assert isinstance(sup, AFCrescent)

        if "ArcL" in shpe:
            sup.curve_l = shpe["ArcL"]
        if "ArcR" in shpe:
            sup.curve_r = shpe["ArcR"]
        return sup

    def _get_svg_path(self) -> Path:
        path = Path([Move(0, 1)])

        if self.curve_l != 0:
            rl = (self.curve_l + 1 / self.curve_l) * 1 / 2
            path.append(
                Arc(
                    Vector2d(rl, rl),
                    0,
                    False,
                    self.curve_l < 0,
                    Vector2d(0, -1),
                )
            )
        else:
            path.append(Line(0, -1))

        if self.curve_r != 0:
            rr = (self.curve_r + 1 / self.curve_r) * 1 / 2
            path.append(
                Arc(
                    Vector2d(rr, rr),
                    0,
                    False,
                    self.curve_r > 0,
                    Vector2d(0, 1),
                )
            )
        else:
            path.append(Line(0, 1))
        path.append(ZoneClose())

        return path


class AFHeart(AFCurveShape):
    def __init__(
        self,
        b_box: BoundingBox,
        design_point: float = 0.1,
    ):
        super().__init__(b_box)
        self.design_point = design_point

    @classmethod
    def from_af(cls, shpe: AFObject, b_box: BoundingBox) -> AFHeart:
        sup = super().from_af(shpe, b_box)
        assert shpe.get_type() == "ShHt", "Not a heart"
        assert isinstance(sup, AFHeart)

        if "Sprd" in shpe:
            sup.design_point = shpe["Sprd"]
        return sup

    def _get_svg_path(self) -> Path:
        bottom = -1 + self.design_point * 2
        return Path(
            [
                Move(0, bottom),
                Curve(0.210521, -1, 0.631586, -1, 0.842101, -0.8),
                Curve(1.05263, -0.6, 1.05263, -0.2, 0.842101, 0.2),
                Curve(0.69473, 0.5, 0.315796, 0.8, 0, 1),
                Curve(-0.315789, 0.8, -0.694735, 0.5, -0.842105, 0.2),
                Curve(-1.05263, -0.2, -1.05263, -0.6, -0.842105, -0.8),
                Curve(-0.631577, -1, -0.210525, -1, 0, bottom),
                ZoneClose(),
            ]
        )


class AFPolygon(AFCurveShape):
    def __init__(
        self,
        b_box: BoundingBox,
        curvature: float = 0,
        sides: int = 5,
        smooth: bool = False,
    ):
        super().__init__(b_box)
        self.curv = curvature
        self.sides = sides
        self.smooth = smooth

    @classmethod
    def from_af(cls, shpe: AFObject, b_box: BoundingBox) -> AFPolygon:
        sup = super().from_af(shpe, b_box)
        assert shpe.get_type() == "ShPy", "Not a polygon"
        assert isinstance(sup, AFPolygon)

        if "Side" in shpe:
            sup.sides = shpe["Side"]
        if "Curv" in shpe:
            sup.curv = shpe["Curv"]
        if "Smth" in shpe:
            sup.smooth = shpe["Smth"]

        return sup

    def _get_svg_path(self) -> Path:
        # The handles of each point are symmetric with respect to the point's
        # connection line to the polygon center.
        # For 100% curvature, the handles coincide with the bezier approximation
        # of a circle, i.e. length of 4/3 tan(pi/(2n)) and perpendicular to the
        # point's symmetry axis.
        # For curvature -> 0+, the handle is placed at 1/3 of the line
        # between the current and the next node.
        # For curvatures in between, the handle position is linearly interpolated
        # between the 0 and 100% positions.
        # For negative curvatures, the 100% handle is mirrored on the straight
        # connection line between the point and its neighbor.

        pt100 = Vector2d(4 / 3 * math.tan(math.pi / (2 * self.sides)), 0)
        pts = points_on_ellipse(
            Vector2d(1, 1),
            Vector2d(0, 0),
            self.sides,
        )

        sidelen = (pts[1] - pts[0]).length

        pt0 = cmath.rect(sidelen * 1 / 3, math.pi / self.sides)

        ptm100 = cmath.rect(pt100.length, 2 * math.pi / self.sides)

        if self.curv > 0:
            if self.smooth:
                hdl = complex(pt100) * self.curv
            else:
                hdl = complex(pt100) * self.curv + pt0 * (1 - self.curv)
        else:
            hdl = ptm100 * (-self.curv) + pt0 * (1 + self.curv)
        hdl_in = -complex(hdl).conjugate()

        path = Path([Move(pts[0])])

        # Draw the polygon
        for i, (pt, pt2) in enumerate(zip(pts, pts[1:] + pts[0:1])):
            angle_incr = cmath.rect(1, 2 * math.pi / self.sides * i)
            angle_incr2 = cmath.rect(1, 2 * math.pi / self.sides * (i + 1))
            if self.curv == 0:
                path.append(Line(pt2))
            else:
                path.append(
                    Curve(pt + hdl * angle_incr, pt2 + hdl_in * angle_incr2, pt2)
                )

        path.append(ZoneClose())
        return path


class AFSegment(AFCurveShape):
    def __init__(
        self,
        b_box: BoundingBox,
        angle: float = 0,
        pos0: float = 0.25,
        pos1: float = 0,
    ):
        super().__init__(b_box)
        self.angle = angle
        self.pos0 = pos0
        self.pos1 = pos1

    @classmethod
    def from_af(cls, shpe: AFObject, b_box: BoundingBox) -> AFSegment:
        sup = super().from_af(shpe, b_box)
        assert shpe.get_type() == "ShSg", "Not a segment"
        assert isinstance(sup, AFSegment)

        if "Angl" in shpe:
            sup.angle = shpe["Angl"]
        if "Pos0" in shpe:
            sup.pos0 = shpe["Pos0"]
        if "Pos1" in shpe:
            sup.pos1 = shpe["Pos1"]

        return sup

    def _get_svg_path(self):
        p = Path()

        # First the unrotated version, then we apply a transform in the end

        pos0 = (0.5 - self.pos0) / 0.5
        pos1 = (self.pos1 - 0.5) / 0.5

        p1 = math.cos(math.asin(pos0)) + pos0 * 1j
        p2 = -math.cos(math.asin(pos1)) - pos1 * 1j

        if pos0 != 1:
            # Start with the first cutoff segment
            p.append(Move(p1))
            p.append(Line(-p1.conjugate()))
            if pos1 == 1:
                p.append(Arc(1 + 1j, 0, True if pos0 > 0 else False, True, p1))
                p.append(ZoneClose())
            else:
                p.append(Arc(1 + 1j, 0, False, True, p2))
        if pos1 != 1:
            if pos0 == 1:
                p.append(Move(p2))
            p.append(Line(-p2.conjugate()))

            if pos0 == 1:
                p.append(Arc(1 + 1j, 0, True if pos1 > 0 else False, True, p2))
            else:
                p.append(Arc(1 + 1j, 0, False, True, p1))
            p.append(ZoneClose())
        if pos0 == 1 and pos1 == 1:
            p.append(move(0, 1))
            p.append(Arc(1 + 1j, 0, True, False, 0 - 1j))
            p.append(Arc(1 + 1j, 0, True, False, 0 + 1j))

        # Now rotate the path
        return p.transform(inkex.Transform(rotate=(90 - self.angle * 180 / math.pi)))


class AFTear(AFCurveShape):
    def __init__(
        self,
        b_box: BoundingBox,
        fixed_rounding: bool = False,
        curv: float = 0.3,
        bend: float = 0,
        tail: float = 0.5,
        ball: float = 0.25,
    ):
        super().__init__(b_box)
        self.fixed_rounding = fixed_rounding
        self.curv = curv
        self.tail = tail
        self.ball = ball
        self.bend = bend

    @classmethod
    def from_af(cls, shpe: AFObject, b_box: BoundingBox) -> AFTear:
        sup = super().from_af(shpe, b_box)
        assert shpe.get_type() == "ShTr", "Not a tear"
        assert isinstance(sup, AFTear)

        if "Fixd" in shpe:
            sup.fixed_rounding = shpe["Fixd"]
        if "Ball" in shpe:
            sup.ball = shpe["Ball"]
        if "Curv" in shpe:
            sup.curv = shpe["Curv"]
        if "Bend" in shpe:
            sup.bend = shpe["Bend"]
        if "Tail" in shpe:
            sup.tail = shpe["Tail"]

        return sup

    def _get_svg_path(self):
        p = Path()
        top = (self.tail * 2 - 1) - 1j
        ball = self.ball if self.fixed_rounding else 1
        right = 1 + 1j * (1 - ball)
        left = -right.conjugate()
        side_handle = -(self.curv * (2 - ball)) * 1j
        topright_handle = self.bend * ((1 - self.tail) + (1 - ball) * 1j)
        topleft_handle = self.bend * (0.7 - self.tail + 1j - 0.5j * ball)

        p.append(Move(top))
        if self.bend != 0 or self.curv != 0:
            p.append(Curve(top + topright_handle, right + side_handle, right))
        else:
            p.append(Line(right))

        p.append(Arc(1 + 1j * (ball), 0, False, True, left))
        if self.bend != 0 or self.curv != 0:
            p.append(Curve(left + side_handle, top + topleft_handle, top))
        else:
            p.append(Line(top))

        p.append(ZoneClose())

        return p


class AFCloud(AFCurveShape):
    def __init__(self, b_box: BoundingBox, sides: int = 12, rint: float = 0.85):
        super().__init__(b_box)
        self.rint = rint
        self.sides = sides

    @classmethod
    def from_af(cls, shpe: AFObject, b_box: BoundingBox) -> AFCloud:
        sup = super().from_af(shpe, b_box)
        assert shpe.get_type() == "ShCl", "Not a cloud"
        assert isinstance(sup, AFCloud)

        if "Bubl" in shpe:
            sup.sides = shpe["Bubl"]
        if "IRad" in shpe:
            sup.rint = shpe["IRad"]
        return sup

    def _get_svg_path(self):
        """Cloud is designed around a critical shape. This shape is an n-sided polygon,
        where semi-circles are glued onto the polyon's sides, and the enveloping circle
        is the outer radius of the shape.

        For an inner radius of 1, all handles have the length l1= (4/3)*tan(pi/(2*2*n))
        For an inner radius of -> 0, the inner handle has length l0= tan(pi/8) * 4/3 ~ 0.5522
        which is the handle length of the classical circle approximation (4 segments)

        Trigonometry tells us that the critical shape corresponds
        to an inner radius of
        1 / ((tan(pi / n) + 1) cos(pi / n))
        and the radius of the inscribed semicircles is rin = tan(pi/n) / (tan(pi / n) + 1)
        so they have a handle length of lcrit=4/3 * tan(pi/8) * rin

        If r > cricical radius:
        - the endpoint of inner handles interpolate linearly between the
          critical one and a handle with length l1 perpendicular to the radius
        - the outer handles are always perpendicular to the radius. The length varies
          linearly between the lcrit and l1
        If r < critical radius:
        - angle of the inner handles is pi/n
        - length of the inner handles decreases linearly from l0 to lcrit
        - length of the outer handle increases linearly from 0 to lcrit"""

        outerpts = points_on_ellipse(
            Vector2d(1, 1),
            Vector2d(0, 0),
            self.sides,
        )
        innerpts = points_on_ellipse(
            Vector2d(self.rint, self.rint),
            Vector2d(0, 0),
            self.sides,
            -math.pi / self.sides,
        )
        rthr = 1 / (
            (math.tan(math.pi / self.sides) + 1) * math.cos(math.pi / self.sides)
        )
        lcrit = (
            math.tan(math.pi / self.sides)
            / (math.tan(math.pi / self.sides) + 1)
            * 4
            / 3
            * math.tan(math.pi / 8)
        )
        l0 = 4 / 3 * math.tan(math.pi / 8)
        l1 = 4 / 3 * math.tan(math.pi / (4 * self.sides))

        # Inner handle
        crit_handle_endpoint = (
            -1j * cmath.rect(rthr, -math.pi / self.sides) - lcrit * 1j
        )
        zero_handle_endpoint = -l0 * 1j
        one_handle_endpoint = (-1j + l1) * cmath.rect(1, -math.pi / self.sides)

        if self.rint < rthr:
            inner_handle_endpoint = (
                crit_handle_endpoint - zero_handle_endpoint
            ) * self.rint / rthr + zero_handle_endpoint
        else:
            inner_handle_endpoint = (one_handle_endpoint - crit_handle_endpoint) / (
                1 - rthr
            ) * (self.rint - rthr) + crit_handle_endpoint

        # Outer handle
        if self.rint < rthr:
            outer_handle_endpoint = -1j - lcrit * self.rint / rthr
        else:
            outer_handle_endpoint = -1j - (
                (l1 - lcrit) / (1 - rthr) * (self.rint - rthr) + lcrit
            )

        outer_handle_right = -outer_handle_endpoint.conjugate()
        inner_handle_right = -inner_handle_endpoint.conjugate()

        path = Path([Move(innerpts[0])])

        for i, (op, ip2) in enumerate(zip(outerpts, innerpts[1:] + innerpts[0:1])):
            angle1 = cmath.rect(1, 2 * math.pi / self.sides * i)

            path.append(
                Curve(
                    inner_handle_endpoint * angle1, outer_handle_endpoint * angle1, op
                )
            )
            path.append(
                Curve(outer_handle_right * angle1, inner_handle_right * angle1, ip2)
            )

        path.append(ZoneClose())

        return path


shape_map: Dict[str, Type[AFShape]] = {
    "ShNR": AFRectangle,
    "ShpE": AFEllipse,
    "ShpT": AFTriangle,
    "ShpD": AFDiamond,
    "ShTz": AFTrapezoid,
    "ShPi": AFPie,
    "ShSS": AFSquareStar,
    "ShDS": AFDoubleStar,
    "ShSt": AFStar,
    "ShCr": AFCrescent,
    "ShHt": AFHeart,
    "ShPy": AFPolygon,
    "ShSg": AFSegment,
    "ShTr": AFTear,
    "ShCl": AFCloud,
}


def parse_shape(
    shpe: AFObject,
    b_box: tuple[float, float, float, float],
    scale: Optional[inkex.Transform] = None,
) -> AFShape:
    shape_type = shpe.get_type()
    if shape_type in shape_map:
        bbox = bbox_helper(b_box)
        if scale:
            xy1 = scale.apply_to_point((bbox.left, bbox.top))
            xy2 = scale.apply_to_point((bbox.right, bbox.bottom))
            bbox = BoundingBox((xy1.x, xy2.x), (xy1.y, xy2.y))
        return shape_map[shape_type].from_af(shpe, bbox)

    raise NotImplementedError(f"Shape type {shape_type} not implemented")



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2020-2022 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""SVG -> XAML Converter classes for Shapes:
<path>, <ellipse>, <circle>, <rect>, <line>, <polygon>, <polyline>"""


import abc
from typing import Union, Optional, Tuple, List, Dict
import math
import copy

import inkex
from inkex.transforms import Transform
from inkex.bezier import bezierslopeatt

from ..xamlobjects import (
    KEY_ATTR,
    XAMLObject,
    Brush,
    SolidColorBrush,
    LinearGradientBrush,
    RadialGradientBrush,
    GradientStop,
    DrawingBrush,
    Number,
    Shape,
    BaseElement,
    DashStyle,
    Pen,
    Canvas,
    GeometryGroup,
    Geometry,
    GeometryDrawing,
    DrawingGroup,
    Drawing,
    Path,
    PathGeometry,
    Rectangle,
    RectangleGeometry,
    Line,
    LineGeometry,
    Ellipse,
    EllipseGeometry,
)
from .utils import convert_attributes
from .config import GLOBAL_SETTINGS
from .base import BaseParser


class ShapeElementParser(BaseParser, abc.ABC):
    """Parse shapes such as rect, circle ect."""

    def __init__(self, svgelement: inkex.ShapeElement):
        super().__init__(svgelement)
        self.effective_style = self.element.specified_style()

    @staticmethod
    def color_to_xaml_color(color: inkex.Color, premultiply_opacity: Number = 1) -> str:
        """Convert a string to an XAML color"""
        xamlcolor = color.to_rgba()
        xamlcolor[3] = int(xamlcolor[3] * 255 * 1 * premultiply_opacity)
        return "#{3:02x}{0:02x}{1:02x}{2:02x}".format(*xamlcolor)

    def fill_gradient(self, svggrad: inkex.Gradient, brush: Brush, opacity: Number = 1):
        """Fills the properties of a gradient"""
        largest = 0
        for stop in svggrad.stops:
            effective_style = stop.specified_style()
            gdstop = GradientStop()
            gdstop_opacity = (
                1
                if effective_style("stop-opacity") is None
                else float(effective_style("stop-opacity"))
            )
            gdstop_color = effective_style("stop-color")
            gdstop.add_property(
                "Color",
                self.color_to_xaml_color(gdstop_color, gdstop_opacity * opacity),
            )
            # fix https://www.w3.org/TR/SVG2/pservers.html#GradientStops
            stop.offset = min(max(stop.offset, largest, 0), 1)
            gdstop.add_property("Offset", stop.offset)
            largest = stop.offset
            brush.add_child(gdstop)
        convert_attributes(
            svggrad,
            brush,
            [["spreadMethod", "SpreadMethod", ["pad", "reflect", "repeat"]]],
        )
        if not GLOBAL_SETTINGS.avalonia:
            convert_attributes(
                svggrad,
                brush,
                [
                    [
                        "gradientUnits",
                        "MappingMode",
                        {
                            "userSpaceOnUse": "Absolute",
                            "objectBoundingBox": "RelativeToBoundingBox",
                        },
                        "RelativeToBoundingBox",
                        "objectBoundingBox",
                    ]
                ],
            )

        def transformer(number):
            # Relative gradient units are not really well transformed in Avalonia,
            # we try what we can do. See also some back and forth in
            # https://github.com/AvaloniaUI/Avalonia/pull/6522
            if (
                GLOBAL_SETTINGS.avalonia
                and svggrad.get("gradientUnits", "objectBoundingBox")
                == "objectBoundingBox"
            ):
                return str(number * 100) + "%"
            return number

        if isinstance(svggrad, inkex.LinearGradient):
            brush.add_properties(
                {
                    "StartPoint": [
                        transformer(svggrad.x1(self.element.root)),
                        transformer(svggrad.y1(self.element.root)),
                    ],
                    "EndPoint": [
                        transformer(svggrad.x2(self.element.root)),
                        transformer(svggrad.y2(self.element.root)),
                    ],
                }
            )
        elif isinstance(svggrad, inkex.RadialGradient):
            self.fill_radial_gradient(svggrad, brush)
        if GLOBAL_SETTINGS.wpf:
            # Fix this once https://github.com/AvaloniaUI/Avalonia/issues/6343 is
            # released
            brush.set_transform(svggrad.gradientTransform)

    def fill_radial_gradient(self, svggrad, brush):
        """Add the equivalent properties of a radial gradient"""
        r = svggrad.r(self.element.root)
        c = [svggrad.cx(self.element.root), svggrad.cy(self.element.root)]
        focus = [svggrad.fx(self.element.root), svggrad.fy(self.element.root)]
        # check if focal point lies outside of circle
        # (https://www.w3.org/TR/SVG11/pservers.html#RadialGradients)
        dist = math.sqrt((c[0] - focus[0]) ** 2 + (c[1] - focus[1]) ** 2)
        if dist > r:
            # in this case place it on the intersection of the circle and the line
            focus = [
                (focus[0] - c[0]) * r * 1 / dist + c[0],
                (focus[1] - c[1]) * r * 1 / dist + c[1],
            ]
        if GLOBAL_SETTINGS.wpf:
            brush.add_properties({"RadiusX": r, "RadiusY": r})
        else:
            if svggrad.get("gradientUnits", "objectBoundingBox") == "userSpaceOnUse":
                # workaround for https://github.com/AvaloniaUI/Avalonia/issues/3995
                # The smaller side of the bounding box of the untransformed element
                # seems to be the basis for the computation of the drawn radius.
                # So first untransform the element
                temp_transform = self.element.transform
                self.element.transform @= -self.element.transform
                # now read the bounding box
                bbox = self.element.bounding_box()
                self.element.transform = temp_transform
                r /= min(bbox.width, bbox.height)
            brush.add_properties({"Radius": r})
        brush.add_properties({"Center": c, "GradientOrigin": focus})

    def get_swatch(self, value, opacity):
        """Convert a gradient to a swatch; and return the property value
        to be set"""
        stop_style = value.href.stops[0].specified_style()
        color = self.color_to_xaml_color(
            stop_style("stop-color"),
            opacity * stop_style("stop-opacity", 1),
        )
        if GLOBAL_SETTINGS.swatches == "color":
            return color
        key = value.href.get_id()
        if GLOBAL_SETTINGS.TOPLEVEL_DICT.get_child(key) is None:
            res = SolidColorBrush()
            res.add_property("Color", color)
            res.add_property(KEY_ATTR, key)
            GLOBAL_SETTINGS.TOPLEVEL_DICT.insert(0, res)
        if GLOBAL_SETTINGS.TOPLEVEL_DICT.get_child(key) is not None:
            return f"{{{GLOBAL_SETTINGS.swatches} {key}}}"
        # mostly for safety reason in case there was an ID collision
        return color

    def postprocess_1d_gradient(
        self, xaml_parent: XAMLObject, value: inkex.Gradient, parsed: XAMLObject
    ) -> Optional[Union[Brush, str]]:
        """Some gradient postprocessing:
        1) Angle between gradient axis and gradient direction
        2) GradientTransform for Canvas"""
        if (  # Avalonia doesn't support DrawingBrush
            GLOBAL_SETTINGS.wpf
            and isinstance(value, inkex.LinearGradient)
            and value.get("gradientUnits", "objectBoundingBox") == "objectBoundingBox"
            # In this case no workaround is needed.
            and (value.x1() != value.x2() and value.y1() != value.y2())
        ):
            # https://www.w3.org/TR/SVG/pservers.html#LinearGradientElementGradientUnitsAttribute
            drwb = DrawingBrush()
            gmd = GeometryDrawing()
            drwb.add_properties(
                {"Viewbox": "0,0,1,1", "Stretch": "Fill", "Drawing": gmd}
            )
            rectgeom = RectangleGeometry()
            rectgeom.add_property("Rect", "0,0,1,1")
            gmd.add_properties({"Brush": parsed, "Geometry": rectgeom})
            return drwb
        if (
            isinstance(value, (inkex.RadialGradient, inkex.LinearGradient))
            and isinstance(xaml_parent, Shape)
            and xaml_parent is not None
            and value.get("gradientUnits") == "userSpaceOnUse"
            # Remove when https://github.com/AvaloniaUI/Avalonia/issues/6343
            # is released
            and not GLOBAL_SETTINGS.avalonia
        ):
            left = xaml_parent.get("Canvas.Left", None)
            top = xaml_parent.get("Canvas.Top", None)
            if left is not None and top is not None:
                # in a canvas, we have to apply a transform to the gradient
                parsed.set_transform(
                    Transform(translate=(-float(left), -float(top)))
                    @ value.gradientTransform
                )
        return parsed

    def get_brush(
        self, xaml_parent: BaseElement, value=None, opacity=1
    ) -> Optional[Union[Brush, str]]:
        """Adds a brush to xaml_parent. Returns true if the brush can be used, else
        false"""
        if value is None:
            try:
                value = self.effective_style("fill")
            except inkex.ColorError:
                value = inkex.Color("black")
            opacity = self.effective_style("fill-opacity")
        parsed: Brush
        if isinstance(value, (inkex.LinearGradient, inkex.RadialGradient)):
            if isinstance(value, inkex.LinearGradient):
                # We can express the swatch as a Resource if the following properties
                # are fulfilled: the inkscape:swatch attribute is set to solid,
                # the element doesn't have an extra fill-opacity (Inkscape doesn't do
                # that, the opacity slider affects the swatch), and there is exactly
                # one stop (i.e. a solid color)
                if (
                    value.href is not None
                    and value.href.get("inkscape:swatch", "None") == "solid"
                    and len(value.href.stops) == 1
                    and opacity == 1
                ):
                    return self.get_swatch(value, opacity)

                parsed = LinearGradientBrush()
            elif isinstance(value, inkex.RadialGradient):
                parsed = RadialGradientBrush()
            self.fill_gradient(value, parsed, opacity)
            return self.postprocess_1d_gradient(xaml_parent, value, parsed)
        if value is None or value == "none":
            return None
        if isinstance(value, inkex.Pattern):
            parsed = self.get_mask_pattern(xaml_parent, value, opacity)

        else:  # probably a color
            # verbose version, probably won't use
            # parsed = SolidColorBrush()
            # xaml_parent.add_property("Brush", parsed)
            # parsed.add_property("Color", '#{3:02x}{0:02x}{1:02x}{2:02x}'
            # .format(*xamlcolor))
            try:
                return self.color_to_xaml_color(inkex.Color(value), opacity)
            except inkex.colors.ColorError:
                print(f"Invalid color value: {self.effective_style['fill']}")
                return None  # resvg-test a-fill-023
        return parsed

    def dashstyle_from_dasharray(
        self, stroke_dasharray: List[float], append=True
    ) -> Optional[Union[DashStyle, str]]:
        """Generates a dashstyle from a given dasharray. In XAML, the dashstyle is given
        as multiples of the stroke thickness, whereas in SVG it's given in user
        units."""
        width = self.effective_style("stroke-width")
        width = inkex.units.convert_unit(width, "px")
        dsty = DashStyle()
        result = ",".join([str(i / width) for i in stroke_dasharray])
        if append:
            dsty.add_property("Dashes", result)
            return dsty
        return result

    def add_pen(self, xaml_parent: BaseElement) -> Optional[Pen]:
        """Add a pen to xaml_parent based on self.element's style"""
        lowlevel = not isinstance(xaml_parent, Shape)
        parsed: BaseElement = xaml_parent
        if lowlevel:
            parsed = Pen()
        stroke = self.effective_style("stroke")
        stroke_opacity = self.effective_style("stroke-opacity")
        result = self.get_brush(parsed, stroke, stroke_opacity)
        if stroke is None:
            return None
        parsed.add_property("Brush" if lowlevel else "Stroke", result)
        linecap_dict = {"butt": "Flat", "round": "Round", "square": "Square"}
        attribute_converter: List[List]
        attribute_converter = [
            [
                "stroke-linejoin",
                "Join"
                if GLOBAL_SETTINGS.avalonia and GLOBAL_SETTINGS.mode == "canvas"
                else "LineJoin",
                {
                    "miter": "Miter",
                    "miter-clip": "Miter",
                    "bevel": "Bevel",
                    "round": "Round",
                },
                "Miter" if GLOBAL_SETTINGS.wpf else "Bevel",
            ],
            ["stroke-width", "Thickness", lambda s: inkex.units.convert_unit(s, "px")],
        ]
        if GLOBAL_SETTINGS.wpf:
            attribute_converter += [
                ["stroke-linecap", "DashCap", linecap_dict, "Square"],
                ["stroke-linecap", "EndLineCap", linecap_dict, "Flat"],
                ["stroke-linecap", "StartLineCap", linecap_dict, "Flat"],
            ]
        else:
            attribute_converter += [["stroke-linecap", "LineCap", linecap_dict, "Flat"]]
        if not lowlevel:
            for lst in attribute_converter:
                lst[1] = "Stroke" + lst[1]
            attribute_converter += [
                [
                    "stroke-dasharray",
                    "StrokeDashArray",
                    lambda s: self.dashstyle_from_dasharray(s, False),
                ]
            ]
        else:
            attribute_converter += [
                ["stroke-dasharray", "DashStyle", self.dashstyle_from_dasharray]
            ]
        # pylint: enable=line-too-long
        convert_attributes(self.effective_style, parsed, attribute_converter)
        if result is not None and lowlevel and isinstance(parsed, Pen):
            return parsed
        return None


class MarkerMixin(ShapeElementParser):
    """Mixin to handle marker conversion logic. Markers don't exist in XAML, so they
    have to be converted into regular objects"""

    @staticmethod
    def get_angle(prev, cspsegment, nxt) -> float:
        """Get the orientation of a marker based on the previous, current and next
        superpath segment"""
        incoming_angle = None
        outgoing_angle = None
        if prev is not None:
            # for lines, the derivative of the bezier parametric curve is not defined at
            # the end points, so we take the derivative very close to the end points
            # Replace this after oo-paths is done
            incoming_angle = inkex.Vector2d(
                bezierslopeatt(prev[1:] + cspsegment[0:2], 1 - 1e-7)
            )
            incoming_angle /= incoming_angle.length + 1e-15
        if nxt is not None:
            outgoing_angle = inkex.Vector2d(
                bezierslopeatt(cspsegment[1:] + nxt[0:2], 0 + 1e-7)
            )
            outgoing_angle /= outgoing_angle.length + 1e-15
        if incoming_angle is not None and outgoing_angle is not None:
            mean_angle = (incoming_angle + outgoing_angle) / 2
        else:
            mean_angle = incoming_angle or outgoing_angle or inkex.Vector2d(1, 0)
        return math.atan2(mean_angle.y, mean_angle.x) * 180 / math.pi

    def find_markers(self):
        """Compile a list of markers (position, angle, Marker reference)
        from the objects' style and nodes"""
        mst = self.effective_style("marker-start")
        mmid = self.effective_style("marker-mid")
        mend = self.effective_style("marker-end")
        path = self.element.get_path().to_superpath()
        nodelist: List[Tuple[Tuple[float, float], float, inkex.Marker]] = []
        # superpaths are stored as Subpath[Segment[Bezier_in, Point, Bezier_out]]
        for subpath in path:
            for segct, segment in enumerate(subpath):
                point = segment[1]
                prev = None if segct == 0 else subpath[segct - 1]
                nxt = None if segct == len(subpath) - 1 else subpath[segct + 1]
                angle = self.get_angle(prev, segment, nxt)
                nodelist.append((point, angle, mmid))
        if nodelist[0]:
            nodelist[0] = tuple(nodelist[0][0:2] + (mst,))  # type: ignore
            if mst is None:
                nodelist = nodelist[1:]
        if nodelist[-1]:
            nodelist[-1] = tuple(nodelist[-1][0:2] + (mend,))  # type: ignore
            if mend is None:
                nodelist = nodelist[:-1]
        return nodelist

    def get_markers(self, lowlevel=True) -> List[Union[DrawingGroup, Canvas]]:
        """Get the markers of an object"""
        # pylint: disable=too-many-locals
        markers = self.find_markers()
        results = []
        for point, angle, marker in markers:
            if marker is None:
                continue
            orient = marker.get("orient", 0)
            if orient != "auto":
                angle = self.element.unittouu(orient)
            # Construct the equivalent transform of the marker (before the clip)
            trs = inkex.Transform(translate=point)
            trs @= inkex.Transform(rotate=angle)

            if marker.get("markerUnits", "strokeWidth") == "strokeWidth":
                trs @= inkex.Transform(
                    scale=self.element.to_dimensionless(
                        self.effective_style("stroke-width")
                    )
                )
            # compute marker scaling, needed for shifting
            vbx = marker.get_viewbox()
            # 3, really (https://www.w3.org/TR/SVG2/painting.html#MarkerWidthAttribute)
            marker_width = self.element.unittouu(marker.get("markerWidth", 3))
            marker_height = self.element.unittouu(marker.get("markerHeight", 3))
            x_scale = marker_width / vbx[2] if vbx[2] > 0 else 1
            y_scale = marker_height / vbx[3] if vbx[3] > 0 else 1
            # check if aspect ratio is preserved
            par = marker.get("preserveAspectRatio", "xMidYMid")
            alignment = par.split()
            if alignment[0] != "none":
                slc = alignment[1] if len(alignment) > 1 else "meet"
                aspect_scale = (
                    min(x_scale, y_scale) if slc == "meet" else max(x_scale, y_scale)
                )
                x_scale = y_scale = aspect_scale
            trs @= inkex.Transform(
                translate=(
                    -self.element.unittouu(marker.get("refX", 0)) * x_scale,
                    -self.element.unittouu(marker.get("refY", 0)) * y_scale,
                )
            )

            trs @= Transform(scale=(x_scale, y_scale))
            from .structural import GroupParser

            parser = GroupParser(marker)
            result: BaseElement = parser.to_xaml_objects() if lowlevel else parser.to_canvas_objects()  # type: ignore # pylint: disable=line-too-long
            results += [result]

            if marker.specified_style()("overflow", default="hidden") in [
                "hidden",
                "scroll",
            ]:  # Construct a clipping path
                mk_width = vbx[2] * x_scale
                mk_height = vbx[3] * y_scale
                x_offset = 0
                y_offset = 0
                if "xMid" in alignment[0]:
                    x_offset -= (marker_width - mk_width) / 2
                elif "xMax" in alignment[0]:
                    x_offset -= marker_width - mk_width
                if "YMid" in alignment[0]:
                    y_offset -= (marker_height - mk_height) / 2
                elif "YMax" in alignment[0]:
                    y_offset -= marker_height - mk_height
                # specify a clipping rectangle on the marker parent
                clip_geometry = RectangleGeometry()
                clip_geometry.add_property(
                    "Rect",
                    ",".join(
                        map(
                            str,
                            [
                                vbx[0],
                                vbx[1],
                                (x_offset + marker_width) / x_scale,
                                (y_offset + marker_height) / x_scale,
                            ],
                        )
                    ),
                )
                result.add_property(
                    "ClipGeometry" if lowlevel else "Clip", clip_geometry
                )

            result.set_transform(trs)

        return results


class NonTextShapeElementParser(ShapeElementParser):
    """Parser for Shapes which are not text elements"""

    @abc.abstractmethod
    def add_attributes(self, geom: Geometry) -> bool:
        """Add functional attributes such as width, height, pathdata... for Geometry
        representation"""

    @abc.abstractmethod
    def add_canvas_attributes(self, geom: Shape) -> bool:
        """Add functional attributes such as width, height, pathdata... for Shape
        (Canvas) representation"""

    @abc.abstractmethod
    def create_geometry(self):
        """Create the geometry child"""

    @abc.abstractmethod
    def get_canvas_object(self):
        """Create the Canvas object child"""

    def needs_wrapper(self, ignore_clip=False) -> bool:
        """Returns true if the shape requires to be wrapped in a DrawingGroup"""
        if (
            (not ignore_clip and self.element.clip is not None)
            or self.effective_style("opacity") != 1
            or self.effective_style("mask") is not None
            or not self.element.transform == inkex.transforms.Transform()
        ):
            return True
        if any(
            self.effective_style(f"marker-{i}") is not None
            for i in ["start", "mid", "end"]
        ):
            return True
        pto = self.paint_order()
        if pto.index("stroke") < pto.index("fill"):
            return True
        return False

    def paint_order(self) -> List[str]:
        """Get sanitized fill order as list"""
        pto: str = self.effective_style("paint-order")
        split = pto.split()
        if "fill" not in split or "markers" not in split or "stroke" not in split:
            split = ["fill", "stroke", "markers"]
        return split

    def to_xaml_objects(self) -> Optional[Drawing]:
        gdr = GeometryDrawing()
        result = gdr
        geometry = self.create_geometry()
        gdr.add_property("Geometry", geometry)
        res = self.add_attributes(geometry)

        gdr.add_property("Brush", self.get_brush(gdr))
        gdr.add_property("Pen", self.add_pen(gdr))
        if self.needs_wrapper():
            parent = DrawingGroup()
            result = parent
            parent.add_name(self.element.get_id())
            parent.set_transform(self.element.transform)
            parent.add_property("ClipGeometry", self.get_clip(parent))
            parent.add_property(
                "OpacityMask", self.get_mask(parent, self.effective_style("mask"))
            )
            parent.add_property("Opacity", self.effective_style("opacity"), 1)
            pto = self.paint_order()
            constituents: Dict[str, List[XAMLObject]] = {"markers": []}
            if isinstance(self, MarkerMixin):
                constituents[
                    "markers"
                ] = self.get_markers()  # pylint: disable=no-member
            if pto.index("fill") == pto.index("stroke") - 1 or (
                pto.index("fill") < pto.index("stroke") and not constituents["markers"]
            ):  #
                # Either fill stroke markers (default), markers fill stroke, or
                # fill markers stroke, but no markers
                constituents["stroke"] = []
                constituents["fill"] = [gdr]
            else:
                strokechild = copy.deepcopy(gdr)
                strokechild.remove_property("Brush")
                fillchild = copy.deepcopy(gdr)
                fillchild.remove_property("Pen")
                constituents.update({"stroke": [strokechild], "fill": [fillchild]})
            for i in pto:
                parent.add_children(constituents[i])

        else:
            gdr.add_name(self.element.get_id())

        if res is not False:
            return result
        return None

    def to_canvas_objects(self) -> Optional[Shape]:
        this = result = fillobj = paintobj = self.get_canvas_object()
        res = self.add_canvas_attributes(this)
        pto = self.paint_order()
        if (
            any(
                self.effective_style(f"marker-{i}") is not None
                for i in ["start", "mid", "end"]
            )
            and isinstance(self, MarkerMixin)
        ) or pto.index("stroke") < pto.index("fill"):
            constituents: Dict[str, List[XAMLObject]] = {"markers": []}
            if isinstance(self, MarkerMixin):
                # pylint: disable=no-member
                constituents["markers"] = self.get_markers(False)
            if pto.index("fill") == pto.index("stroke") - 1 or (
                pto.index("fill") < pto.index("stroke") and not constituents["markers"]
            ):
                # Either fill stroke markers (default), markers fill stroke, or
                # fill markers stroke, but no markers
                constituents["stroke"] = []
                constituents["fill"] = [fillobj]
            else:
                fillobj = copy.deepcopy(paintobj)
                if isinstance(self, TopLeftObjectParser):
                    # In this case, we need to recompute the paint object without stroke
                    # (coordinates change)
                    fillobj = self.get_canvas_object()
                    stroke = self.effective_style.get("stroke-width", None)
                    self.effective_style["stroke-width"] = 0
                    self.add_canvas_attributes(fillobj)
                    self.effective_style["stroke-width"] = stroke
                constituents.update({"stroke": [paintobj], "fill": [fillobj]})
            if sum(1 if len(i) > 0 else 0 for i in constituents.values()) > 1:
                parent = Canvas()
                for i in pto:
                    parent.add_children(constituents[i])
                result = parent
            # we copy the transform up. This is legal because transforms take the
            # stroke width into account only for rects and ellipses, for which
            # markers are invalid
            result.add_property("RenderTransform", this.get("RenderTransform", None))
            this.remove_property("RenderTransform")
        blur = self.get_blur()
        if GLOBAL_SETTINGS.wpf and blur is not None:
            result.add_property("Effect", blur)
        result.add_property("Clip", self.get_clip(this))
        result.add_property(
            "OpacityMask", self.get_mask(this, self.effective_style("mask"))
        )
        result.add_property("Opacity", self.effective_style("opacity"), 1)
        fillobj.add_property("Fill", self.get_brush(fillobj))
        paintobj.add_property("Pen", self.add_pen(paintobj))
        result.add_name(self.element.get_id())

        if res is not False:
            return result
        return None

    def to_clippath(self) -> Optional[Geometry]:
        """Parse the object into a clippath"""
        gdr = self.create_geometry()
        result = gdr
        if self.needs_wrapper(ignore_clip=True):  # no nested clippaths
            parent = GeometryGroup()
            result = parent
            parent.set_transform(self.element.transform)
            parent.add_name(self.element.get_id())
            parent.add_child(gdr)
        else:
            gdr.add_name(self.element.get_id())
        res = self.add_attributes(gdr)
        if (
            self.effective_style("visibility") == "hidden"
            or self.effective_style("display") == "none"
        ):  # https://www.w3.org/TR/SVG11/masking.html#EstablishingANewClippingPath
            return None
        if res is not False:
            return result
        return None


class PathParser(NonTextShapeElementParser, MarkerMixin):
    """Parser for <path> elements"""

    def __init__(self, svgelement: inkex.PathElement):
        super().__init__(svgelement)

    def create_geometry(self) -> PathGeometry:
        return PathGeometry()

    def get_canvas_object(self):
        return Path()

    def add_attributes(self, geom: Geometry):
        geom.add_property("Figures", str(self.element.path))
        convert_attributes(
            self.effective_style,
            geom,
            [
                [
                    "fill-rule",
                    "FillRule",
                    {
                        "evenodd": "EvenOdd",
                        "nonzero": "Nonzero" if GLOBAL_SETTINGS.wpf else "NonZero",
                    },
                    "EvenOdd",
                ]
            ],
        )
        return geom

    def add_canvas_attributes(self, geom: Shape):
        data = PathGeometry()
        geom.add_property("Data", data)
        geom.set_transform(self.element.transform)
        self.add_attributes(data)

    def to_clippath(self) -> Optional[Geometry]:
        result = super().to_clippath()
        if result is not None:
            convert_attributes(
                self.effective_style,
                result,
                [
                    [
                        "clip-rule",
                        "FillRule",
                        {
                            "evenodd": "EvenOdd",
                            "nonzero": "Nonzero" if GLOBAL_SETTINGS.wpf else "NonZero",
                        },
                        "EvenOdd",
                    ]
                ],
            )
        return result


class PolygonPolylineParser(PathParser):
    """Parser for polygon and polyline elements (not created by Inkscape). They
    are simply converted to paths."""

    def add_attributes(self, geom: Geometry):
        path = str(self.element.get_path())
        geom.add_property("Figures", path)
        return geom


class LineParser(NonTextShapeElementParser, MarkerMixin):
    """Parser for <line> elements"""

    def __init__(self, svgelement: inkex.Line):
        super().__init__(svgelement)

    def create_geometry(self) -> LineGeometry:
        return LineGeometry()

    def get_canvas_object(self):
        return Line()

    def get_xy(self) -> List:
        """Return a list [x1, y1, x2, y2]"""
        return [self.element.x1, self.element.y1, self.element.x2, self.element.y2]

    def add_attributes(self, geom: Geometry):
        x_y = self.get_xy()
        geom.add_properties({"StartPoint": x_y[0:2], "EndPoint": x_y[2:4]})
        return geom

    def add_canvas_attributes(self, geom: Shape):
        geom.add_properties(dict(zip(["X1", "Y1", "X2", "Y2"], self.get_xy())))
        geom.set_transform(self.element.transform)


class TopLeftObjectParser(NonTextShapeElementParser):
    """Parse to elements whose position is defined by Canvas.Left and Canvas.Height"""

    def __init__(self, svgelement: Union[inkex.Rectangle, inkex.Circle]):
        super().__init__(svgelement)

    @abc.abstractmethod
    def topleftwidthheight(self) -> Tuple:
        """Get the top, left, width and height of a shape; taking stroke into account"""

    def add_canvas_attributes(self, geom: Shape):
        stroke = (
            0
            if self.effective_style("stroke") is None
            else self.element.to_dimensionless(self.effective_style("stroke-width"))
        )
        top, left, width, height = self.topleftwidthheight()
        left = left - stroke / 2
        top = top - stroke / 2
        attrs = {
            "Canvas.Left": left,
            "Canvas.Top": top,
            "Width": width + stroke,
            "Height": height + stroke,
        }
        effective_transform = (
            Transform(translate=(-left, -top))
            @ self.element.transform
            @ Transform(translate=(left, top))
        )
        geom.set_transform(effective_transform)
        geom.add_properties(attrs)


class RectangleParser(TopLeftObjectParser):
    """Parse a <rect> element"""

    def __init__(self, svgelement: inkex.Rectangle):
        super().__init__(svgelement)

    def create_geometry(self) -> RectangleGeometry:
        return RectangleGeometry()

    def get_canvas_object(self):
        return Rectangle()

    def topleftwidthheight(self) -> Tuple:
        return (
            self.element.top,
            self.element.left,
            self.element.width,
            self.element.height,
        )

    def add_attributes(self, geom: Geometry):
        geom.add_property(
            "Rect",
            ",".join(
                map(
                    str,
                    [
                        self.element.left,
                        self.element.top,
                        self.element.width,
                        self.element.height,
                    ],
                )
            ),
        )
        if self.element.rx > 0:
            geom.add_property("RadiusX", self.element.rx)
        if self.element.ry > 0:
            geom.add_property("RadiusY", self.element.ry)
        if self.element.width == 0 or self.element.height == 0:
            # lowlevel attributes don't have visibility property. Ignore element
            return False
        return geom

    def add_canvas_attributes(self, geom: Shape):
        if not GLOBAL_SETTINGS.avalonia:
            # This has been added in https://github.com/AvaloniaUI/Avalonia/pull/7914,
            # and will be part of 0.11
            geom.add_properties(
                {"RadiusX": self.element.rx, "RadiusY": self.element.ry}
            )
        super().add_canvas_attributes(geom)
        if self.element.width == 0 or self.element.height == 0:
            geom.set_invisible()


class EllipseParser(TopLeftObjectParser):
    """Parse an <ellipse> or <circle> element"""

    def __init__(self, svgelement: inkex.elements._polygons.EllipseBase):
        super().__init__(svgelement)

    def create_geometry(self) -> EllipseGeometry:
        return EllipseGeometry()

    def get_canvas_object(self):
        return Ellipse()

    def topleftwidthheight(self) -> Tuple:
        return (
            self.element.center[1] - self.element.rxry()[1],
            self.element.center[0] - self.element.rxry()[0],
            2 * self.element.rxry()[0],
            2 * self.element.rxry()[1],
        )

    def add_attributes(self, geom: Geometry):
        attrs = {
            "RadiusX": self.element.rxry()[0],
            "RadiusY": self.element.rxry()[1],
            "Center": ",".join(map(str, self.element.center)),
        }
        geom.add_properties(attrs)
        if self.element.rxry()[0] == 0 or self.element.rxry()[1] == 0:
            # lowlevel attributes don't have visibility property. Ignore element
            return False
        return geom

    def add_canvas_attributes(self, geom: Shape):
        super().add_canvas_attributes(geom)
        if self.element.rxry()[0] == 0 or self.element.rxry()[1] == 0:
            geom.set_invisible()
        return geom



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (c) 2024 jonathan.neuhauser@outlook.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
Enums and data structures for CGM files
"""

from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Union

Enum.__repr__ = lambda self: self.__class__.__name__ + "." + self.name  # type: ignore


class RealTypeEnum(Enum):
    FLOATING_POINT = 0
    FIXED_POINT = 1


class ScalingModeEnum(Enum):
    ABSTRACT = 0
    METRIC = 1


class VDCTypeEnum(Enum):
    INTEGER = 0
    REAL = 1


class ColourSelectionModeEnum(Enum):
    INDEXED = 0
    DIRECT = 1


class WidthSpecificationModeEnum(Enum):
    ABSOLUTE = 0
    SCALED = 1
    FRACTIONAL = 2
    MM = 3


class InteriorStyleEnum(Enum):
    HOLLOW = 0
    SOLID = 1
    PATTERN = 2
    HATCH = 3
    EMPTY = 4
    GEOMETRIC_PATTERN = 5
    INTERPOLATED = 6


class ColourModelEnum(Enum):
    RGB = 1
    CIELAB = 2
    CIELUV = 3
    CMYK = 4
    RGB_RELATED = 5


@dataclass
class Point:
    x: Union[float, int]
    y: Union[float, int]

    def __complex__(self):
        return self.x + self.y * 1j


class Colour:
    pass


@dataclass
class DirectColour(Colour):
    v1: int
    v2: int
    v3: int
    v4: Optional[int] = field(default=0)
    colour_model: ColourModelEnum = field(default=ColourModelEnum.RGB, init=False)


@dataclass
class IndexColour(Colour):
    index: int


class PolygonSetLineTypeEnum(Enum):
    INVISIBLE = 0
    VISIBLE = 1
    CLOSE_INVISIBLE = 2
    CLOSE_VISIBLE = 3


@dataclass
class PolygonSetEntry:
    pt: Point
    linetype: PolygonSetLineTypeEnum


class ClipIndicatorEnum(Enum):
    OFF = 0
    ON = 1


class TextPrecisionEnum(Enum):
    STRING = 0
    CHARACTER = 1
    STROKE = 2


class TextFinalFlag(Enum):
    NOT_FINAL = 0
    FINAL = 1


class TextHorizontalAlignmentEnum(Enum):
    NORMAL_HORIZONTAL = 0
    LEFT = 1
    CENTER = 2
    RIGHT = 3
    CONTINUOUS_HORIZONTAL = 4


class TextVerticalAlignmentEnum(Enum):
    NORMAL_VERTICAL = 0
    TOP = 1
    CAP = 2
    HALF = 3
    BASE = 4
    BOTTOM = 5
    CONTINUOUS_VERTICAL = 6


class TextPathEnum(Enum):
    RIGHT = 0
    LEFT = 1
    UP = 2
    DOWN = 3


class LineTypeEnum(Enum):
    SOLID = 1
    DASH = 2
    DOT = 3
    DASH_DOT = 4
    DASH_DOT_DOT = 5
    PRIVATE = -1

    @classmethod
    def _missing_(cls, value):
        return cls.PRIVATE


class EdgeVisibilityEnum(Enum):
    OFF = 0
    ON = 1


class CharacterSetTypeEnum(Enum):
    G_SET_94 = 0
    G_SET_96 = 1
    MULTIBYTE_G_SET_94 = 2
    MULTIBYTE_G_SET_96 = 3
    COMPLETE = 4


class CharacterCodingEnum(Enum):
    BASIC_7BIT = 0
    BASIC_8BIT = 1
    EXTENDED_7BIT = 2
    EXTENDED_8BIT = 3


class ArcClosureEnum(Enum):
    PIE = 0
    CHORD = 1


class TransparencyEnum(Enum):
    OFF = 0
    ON = 1


class ASFTypeEnum(Enum):
    LINE_TYPE_ASF = 0
    LINE_WIDTH_ASF = 1
    LINE_COLOUR_ASF = 2
    MARKER_TYPE_ASF = 3
    MARKER_SIZE_ASF = 4
    MARKER_COLOUR_ASF = 5
    TEXT_FONT_INDEX_ASF = 6
    TEXT_PRECISION_ASF = 7
    CHARACTER_EXPANSION_FACTOR_ASF = 8
    CHARACTER_SPACING_ASF = 9
    TEXT_COLOUR_ASF = 10
    INTERIOR_STYLE_ASF = 11
    FILL_COLOUR_ASF = 12
    HATCH_INDEX_ASF = 13
    PATTERN_INDEX_ASF = 14
    EDGE_TYPE_ASF = 15
    EDGE_WIDTH_ASF = 16
    EDGE_COLOUR_ASF = 17


class ASFValue(Enum):
    INDIVIDUAL = 0
    BUNDLED = 1


class MarkerTypeEnum(Enum):
    DOT = 1
    PLUS = 2
    ASTERISK = 3
    CIRCLE = 4
    CROSS = 5
    PRIVATE = -1

    @classmethod
    def _missing_(cls, value):
        return cls.PRIVATE


class PolybezierContinuityEnum(Enum):
    DISCONTINUOUS = 1
    CONTINUOUS = 2


class CapIndicatorEnum(Enum):
    UNSPECIFIED = 1
    BUTT = 2
    ROUND = 3
    PROJECTING_SQUARE = 4
    TRIANGLE = 5


class DashCapIndicatorEnum(Enum):
    UNSPECIFIED = 1
    BUTT = 2
    MATCH = 3


class RestrictionTypeEnum(Enum):
    BASIC = 1
    BOXED_CAP = 2
    BOXED_ALL = 3
    ISOTROPIC_CAP = 4
    ISOTROPIC_ALL = 5
    JUSTIFIED = 6



# coding=utf-8
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# pylint: disable=invalid-name,missing-docstring
"""Deprecated ffgeom API"""

from collections import namedtuple

from inkex.deprecated import deprecate
from inkex.transforms import DirectedLineSegment as NewSeg

try:
    NaN = float("NaN")
except ValueError:
    PosInf = 1e300000
    NaN = PosInf / PosInf


class Point(namedtuple("Point", "x y")):
    __slots__ = ()

    def __getitem__(self, key):
        if isinstance(key, str):
            key = "xy".index(key)
        return super(Point, self).__getitem__(key)


class Segment(NewSeg):
    @deprecate
    def __init__(self, e0, e1):
        """inkex.transforms.DirectedLineSegment((x1, y1), (x2, y2))"""
        if isinstance(e0, dict):
            e0 = (e0["x"], e0["y"])
        if isinstance(e1, dict):
            e1 = (e1["x"], e1["y"])
        super(Segment, self).__init__(e0, e1)

    def __getitem__(self, key):
        if key:
            return {"x": self.x.maximum, "y": self.y.maximum}
        return {"x": self.x.minimum, "y": self.y.minimum}

    delta_x = lambda self: self.width
    delta_y = lambda self: self.height
    run = delta_x
    rise = delta_y

    def distanceToPoint(self, p):
        return self.distance_to_point(p["x"], p["y"])

    def perpDistanceToPoint(self, p):
        return self.perp_distance(p["x"], p["y"])

    def angle(self):
        return super(Segment, self).angle

    def length(self):
        return super(Segment, self).length

    def pointAtLength(self, length):
        return self.point_at_length(length)

    def pointAtRatio(self, ratio):
        return self.point_at_ratio(ratio)

    def createParallel(self, p):
        self.parallel(p["x"], p["y"])


@deprecate
def intersectSegments(s1, s2):
    """transforms.Segment(s1).intersect(s2)"""
    return Point(*s1.intersect(s2))


@deprecate
def dot(s1, s2):
    """transforms.Segment(s1).dot(s2)"""
    return s1.dot(s2)



#!/usr/bin/env python3

# Copyright (C) 2024 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Takes the parsed object tree, and applies certain enhancements
to it to make it easier to convert.
"""

from __future__ import annotations

from typing import Any, Optional

from .sharedaf import (
    SharedAFDictObject,
    SharedAFListObject,
    SharedAFObject,
    SharedDocument,
)
from .types import AFDictObject, AFListObject, Document, Field, UnknownStruct
from .utils import BinaryParser


def process_obj(
    value: Any, field: Field, doc_id: Optional[int] = None, tag: Optional[str] = None
) -> Any:
    if isinstance(value, list):
        # this was an array
        return [process_obj(o, field, doc_id, tag) for o in value]

    if isinstance(value, AFDictObject):
        return process_dictobj(value, doc_id, tag)
    if isinstance(value, AFListObject):
        return process_listobj(value, doc_id, tag)
    if isinstance(value, UnknownStruct):
        assert field.parent is not None
        return process_struct(value, tag, field.parent.get_type())
    return value


def process_dictobj(
    obj: AFDictObject, doc_id: Optional[int] = None, tag: Optional[str] = None
) -> SharedAFDictObject:
    doc_id = SharedAFObject.new_doc() if doc_id is None else doc_id

    shared = (
        SharedDocument(
            id=obj.id,
            types=obj.types,
            status=obj.status,
            fields=obj.fields,
            doc_format=obj.doc_format,
            docid=doc_id,
        )
        if isinstance(obj, Document)
        else SharedAFDictObject(
            id=obj.id,
            types=obj.types,
            status=obj.status,
            fields=obj.fields,
            docid=doc_id,
        )
    )

    for tag, field in shared.fields.items():
        if isinstance(field, list):
            # Duplicate keys, process them one by one
            field[:] = [
                Field(o.type, process_obj(o.value, o, doc_id, tag), shared)
                for o in field
            ]
            continue
        field.value = process_obj(field.value, field, doc_id, tag)
        field.parent = shared

    return shared


def process_listobj(
    obj: AFListObject, doc_id: Optional[int] = None, tag: Optional[str] = None
) -> SharedAFListObject:
    doc_id = SharedAFObject.new_doc() if doc_id is None else doc_id

    shared = SharedAFListObject(
        id=obj.id, types=obj.types, status=obj.status, fields=obj.fields, docid=doc_id
    )
    for field in shared.fields:
        field.value = process_obj(field.value, field, doc_id)
        field.parent = shared
    return shared


def process_struct(
    value: UnknownStruct,
    field_tag: Optional[str] = None,
    parent_tag: Optional[str] = None,
) -> Any:
    size = len(value.data)
    parser = BinaryParser(value.data)
    if size == 4 and parent_tag == "BGrd" and field_tag == "Colr":
        return tuple(parser.read_u8() for _ in range(4))
    elif field_tag == "_col":
        if size == 8:
            if parent_tag == "LABA":
                return tuple(parser.read_u16() for _ in range(4))
            if parent_tag == "GRAY":
                return (parser.read_float(), parser.read_float())
        if size == 16:
            if parent_tag in ["HSLA", "RGBA"]:
                return tuple(parser.read_float() for _ in range(4))
        if size == 20:
            if parent_tag == "CMYK":
                return tuple(parser.read_float() for _ in range(5))
        raise ValueError("Unknown color format")
    return value


def process(doc: Document) -> SharedDocument:
    return process_dictobj(doc, SharedAFObject.new_doc(), None)  # type: ignore



#
# Copyright 2011-2022 Martin Owens <doctormo@geek-2.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""
Wraps the gtk treeview and iconview in something a little nicer.
"""

import logging

from typing import Tuple, Type, Optional
from gi.repository import Gtk, Gdk, GObject, GdkPixbuf, Pango

from .pixmap import PixmapManager, SizeFilter

GOBJ = GObject.TYPE_PYOBJECT


def default(item, attr, d=None):
    """Python logic to choose an attribute, call it if required and return"""
    if hasattr(item, attr):
        prop = getattr(item, attr)
        if callable(prop):
            prop = prop()
        return prop
    return d


def cmp(a, b):
    """Compare two objects"""
    return (a > b) - (a < b)


def item_property(name, d=None):
    def inside(item):
        return default(item, name, d)

    return inside


def label(obj):
    if isinstance(obj, tuple):
        return " or ".join([label(o) for o in obj])
    if not isinstance(obj, type):
        obj = type(obj)
    return obj.__name__


class BaseView:
    """Controls for tree and icon views, a base class"""

    widget_type: Optional[Type[Gtk.Widget]] = None

    def __init__(self, widget, liststore=None, **kwargs):
        if not isinstance(widget, self.widget_type):
            lbl1 = label(self.widget_type)
            lbl2 = label(widget)
            raise TypeError(f"Wrong widget type: Expected {lbl1} got {lbl2}")

        self.selected_signal = kwargs.get("selected", None)
        self._iids = []
        self._list = widget
        self.args = kwargs
        self.selected = None
        self._data = None
        self.no_dupes = True
        self._model = self.create_model(liststore or widget.get_model())
        self._list.set_model(self._model)
        self.setup()

        self._list.connect(self.changed_signal, self.item_selected_signal)

    def get_model(self):
        """Returns the current data store model"""
        return self._model

    def create_model(self, liststore):
        """Setup the model and list"""
        if not isinstance(liststore, (Gtk.ListStore, Gtk.TreeStore)):
            lbl = label(liststore)
            raise TypeError(f"Expected List or TreeStore, got {lbl}")
        return liststore

    def refresh(self):
        """Attempt to refresh the listview"""
        self._list.queue_draw()

    def setup(self):
        """Setup columns, views, sorting etc"""
        pass

    def get_item_id(self, item):
        """
        Return an id set against this item.

        If item.get_id() is set then duplicates will be ignored.
        """
        if hasattr(item, "get_id"):
            return item.get_id()
        return None

    def replace(self, new_item, item_iter=None):
        """Replace all items, or a single item with object"""
        if item_iter:
            self.remove_item(item_iter)
            self.add_item(new_item)
        else:
            self.clear()
            self._data = new_item
            self.add_item(new_item)

    def item_selected(self, item=None, *others):
        """Base method result, called as an item is selected"""
        if self.selected != item:
            self.selected = item
            if self.selected_signal and item:
                self.selected_signal(item)

    def remove_item(self, item=None):
        """Remove an item from this view"""
        return self._model.remove(self.get_iter(item))

    def check_item_id(self, item):
        """Item id is recorded to guard against duplicates"""
        iid = self.get_item_id(item)
        if iid in self._iids and self.no_dupes:
            raise ValueError(f"Will not add duplicate row {iid}")
        if iid:
            self._iids.append(iid)

    def __iter__(self):
        ret = []

        def collect_all(store, treepath, treeiter):
            ret.append((self.get_item(treeiter), treepath, treeiter))

        self._model.foreach(collect_all)
        return ret.__iter__()

    def set_sensitive(self, sen=True):
        """Proxy the GTK property for sensitivity"""
        self._list.set_sensitive(sen)

    def clear(self):
        """Clear all items from this treeview"""
        self._iids = []
        self._model.clear()

    def item_double_clicked(self, *items):
        """What happens when you double click an item"""
        return items  # Nothing

    def get_item(self, item_iter):
        """Return the object of attention from an iter"""
        return self._model[self.get_iter(item_iter)][0]

    def get_iter(self, item, path=False):
        """Return the iter given the item"""
        if isinstance(item, Gtk.TreePath):
            return item if path else self._model.get_iter(item)
        if isinstance(item, Gtk.TreeIter):
            return self._model.get_path(item) if path else item
        for src_item, src_path, src_iter in self:
            if item == src_item:
                return src_path if path else src_iter
        return None


class TreeView(BaseView):
    """Controls and operates a tree view."""

    column_size = 16
    widget_type = Gtk.TreeView
    changed_signal = "cursor_changed"

    def setup(self):
        """Setup the treeview"""
        self._sel = self._list.get_selection()
        self._sel.set_mode(Gtk.SelectionMode.MULTIPLE)
        self._list.connect("button-press-event", self.item_selected_signal)
        # Separators should do something
        self._list.set_row_separator_func(TreeView.is_separator, None)
        super().setup()

    @staticmethod
    def is_separator(model, item_iter, data):
        """Internal function for seperator checking"""
        return isinstance(model.get_value(item_iter, 0), Separator)

    def get_selected_items(self):
        """Return a list of selected item objects"""
        return [self.get_item(row) for row in self._sel.get_selected_rows()[1]]

    def set_selected_items(self, *items):
        """Select the given items"""
        self._sel.unselect_all()
        for item in items:
            path_item = self.get_iter(item, path=True)
            if path_item is not None:
                self._sel.select_path(path_item)

    def is_selected(self, item):
        """Return true if the item is selected"""
        return self._sel.iter_is_selected(self.get_iter(item))

    def add(self, target, parent=None):
        """Add all items from the target to the treeview"""
        for item in target:
            self.add_item(item, parent=parent)

    def add_item(self, item, parent=None):
        """Add a single item image to the control, returns the TreePath"""
        if item is not None:
            self.check_item_id(item)
            return self._add_item([item], self.get_iter(parent))
        raise ValueError("Item can not be None.")

    def _add_item(self, item, parent):
        return self.get_iter(self._model.append(parent, item), path=True)

    def item_selected_signal(self, *args, **kwargs):
        """Signal for selecting an item"""
        return self.item_selected(*self.get_selected_items())

    def item_button_clicked(self, _, event):
        """Signal for mouse button click"""
        if event is None or event.type == Gdk.EventType._2BUTTON_PRESS:
            self.item_double_clicked(*self.get_selected_items())

    def expand_item(self, item, expand=True):
        """Expand one of our nodes"""
        self._list.expand_row(self.get_iter(item, path=True), expand)

    def create_model(self, liststore=None):
        """Set up an icon view for showing gallery images"""
        if liststore is None:
            liststore = Gtk.TreeStore(GOBJ)
        return super().create_model(liststore)

    def create_column(self, name, expand=True):
        """
        Create and pack a new column to this list.

        Args:
            name: Label in the column header
            expand: Should the column expand
        """
        return ViewColumn(self._list, name, expand=expand)

    def create_sort(self, *args, **kwargs):
        """
        Create and attach a sorting view to this list.

        see :class:`ViewSort` arguments for details.
        """
        return ViewSort(self._list, *args, **kwargs)


class ComboBox(TreeView):
    """Controls and operates a combo box list."""

    widget_type = Gtk.ComboBox
    changed_signal = "changed"

    def setup(self):
        pass

    def get_selected_item(self):
        """Return the selected item of this combo box"""
        return self.get_item(self._list.get_active_iter())

    def set_selected_item(self, item):
        """Set the given item as the selected item"""
        self._list.set_active_iter(self.get_iter(item))

    def is_selected(self, item):
        """Returns true if this item is the selected item"""
        return self.get_selected_item() == item

    def get_selected_items(self):
        """Return a list of selected items (one)"""
        return [self.get_selected_item()]


class IconView(BaseView):
    """Allows a simpler IconView for DBus List Objects"""

    widget_type = Gtk.IconView
    changed_signal = "selection-changed"

    def __init__(self, widget, pixmaps, *args, **kwargs):
        super().__init__(widget, *args, **kwargs)
        self.pixmaps = pixmaps

    def set_selected_item(self, item):
        """Sets the selected item to this item"""
        path = self.get_iter(item, path=True)
        if path:
            self._list.set_cursor(path, None, False)

    def get_selected_items(self):
        """Return the seleced item"""
        return [self.get_item(path) for path in self._list.get_selected_items()]

    def create_model(self, liststore):
        """Setup the icon view control and model"""
        if not liststore:
            liststore = Gtk.ListStore(GOBJ, str, GdkPixbuf.Pixbuf)
        return super().create_model(liststore)

    def setup(self):
        """Setup the columns for the iconview"""
        self._list.set_markup_column(1)
        self._list.set_pixbuf_column(2)
        super().setup()

    def add(self, target):
        """Add all items from the target to the iconview"""
        for item in target:
            self.add_item(item)

    def add_item(self, item):
        """Add a single item image to the control"""
        if item is not None:
            self.check_item_id(item)
            return self._add_item(item)
        raise ValueError("Item can not be None.")

    def get_markup(self, item):
        """Default text return for markup."""
        return default(item, "name", str(item))

    def get_icon(self, item):
        """Default icon return, pixbuf or gnome theme name"""
        return default(item, "icon", None)

    def _get_icon(self, item):
        return self.pixmaps.get(self.get_icon(item), item=item)

    def _add_item(self, item):
        """
        Each item's properties must be stuffed into the ListStore directly
        or the IconView won't see them, but only if on auto.
        """
        if not isinstance(item, (tuple, list)):
            item = [item, self.get_markup(item), self._get_icon(item)]
        return self._model.append(item)

    def item_selected_signal(self, *args, **kwargs):
        """Item has been selected"""
        return self.item_selected(*self.get_selected_items())


class ViewSort(object):
    """
    A sorting function for use is ListViews

     ascending - Boolean which direction to sort
     contains - Contains this string
     data - A string or function to get data from each item.
     exact - Compare to this exact string instead.
    """

    def __init__(self, widget, data=None, ascending=False, exact=None, contains=None):
        self.tree = None
        self.data = data
        self.asc = ascending
        self.comp = exact.lower() if exact else None
        self.cont = contains
        self.tree = widget
        self.resort()

    def get_data(self, model, list_iter):
        """Generate sortable data from the item"""
        item = model.get_value(list_iter, 0)
        if isinstance(self.data, str):
            value = getattr(item, self.data)
        elif callable(self.data):
            value = self.data(item)
        return value

    def sort_func(self, model, iter1, iter2, data):
        """Called by Gtk to sort items"""
        value1 = self.get_data(model, iter1)
        value2 = self.get_data(model, iter2)
        if value1 == None or value2 == None:
            return 0
        if self.comp:
            if cmp(self.comp, value1.lower()) == 0:
                return 1
            elif cmp(self.comp, value2.lower()) == 0:
                return -1
            return 0
        elif self.cont:
            if self.cont in value1.lower():
                return 1
            elif self.cont in value2.lower():
                return -1
            return 0
        if value1 < value2:
            return 1
        if value2 < value1:
            return -1
        return 0

    def resort(self):
        model = self.tree.get_model()
        model.set_sort_func(0, self.sort_func, None)
        if self.asc:
            model.set_sort_column_id(0, Gtk.SortType.ASCENDING)
        else:
            model.set_sort_column_id(0, Gtk.SortType.DESCENDING)


class ViewColumn(object):
    """Add a column to a gtk treeview.

    Args:
        name: The column name used as a label.
        expand: Set column expansion.
    """

    def __init__(self, widget, name, expand=False):
        if isinstance(widget, Gtk.TreeView):
            column = Gtk.TreeViewColumn((name))
            column.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
            column.set_expand(expand)
            self._column = column
            widget.append_column(self._column)
        else:
            # Deal with possible drop down lists
            self._column = widget

    def add_renderer(self, renderer, func, expand=True):
        """Set a custom renderer"""
        self._column.pack_start(renderer, expand)
        self._column.set_cell_data_func(renderer, func, None)
        return renderer

    def add_image_renderer(self, icon, pad=0, pixmaps=None, size=None):
        """Set the image renderer

        Args:
            icon: The function that returns the image to be displayed.
            pad: The amount of padding around the image.
            pixmaps: The pixmap manager to use to get images.
            size: Restrict the images to this size.
        """
        # Manager where icons will be pulled from
        filters = [SizeFilter] if size else []
        pixmaps = pixmaps or PixmapManager(
            "", pixmap_dir="./", filters=filters, size=size
        )

        renderer = Gtk.CellRendererPixbuf()
        renderer.set_property("ypad", pad)
        renderer.set_property("xpad", pad)
        func = self.image_func(icon or self.default_icon, pixmaps)
        return self.add_renderer(renderer, func, expand=False)

    def add_text_renderer(self, text, wrap=None, template=None):
        """Set the text renderer.

        Args:
            text: The function that returns the text to be displayed.
            wrap: The wrapping setting for this renderer.
            template: A standard template used for this text markup.
        """
        renderer = Gtk.CellRendererText()
        if wrap is not None:
            renderer.props.wrap_width = wrap
            renderer.props.wrap_mode = Pango.WrapMode.WORD

        renderer.props.background_set = True
        renderer.props.foreground_set = True

        func = self.text_func(text or self.default_text, template)
        return self.add_renderer(renderer, func, expand=True)

    @classmethod
    def clean(cls, text, markup=False):
        """Clean text of any pango markup confusing chars"""
        if text is None:
            text = ""
        if isinstance(text, (str, int, float)):
            if markup:
                text = str(text).replace("<", "&lt;").replace(">", "&gt;")
            return str(text).replace("&", "&amp;")
        elif isinstance(text, dict):
            return dict([(k, cls.clean(v)) for k, v in text.items()])
        elif isinstance(text, (list, tuple)):
            return tuple([cls.clean(value) for value in text])
        raise TypeError("Unknown value type for text: %s" % str(type(text)))

    def get_callout(self, call, default=None):
        """Returns the right kind of method"""
        if isinstance(call, str):
            call = item_property(call, default)
        return call

    def text_func(self, call, template=None):
        """Wrap up our text functionality"""
        callout = self.get_callout(call)

        def internal(column, cell, model, item_iter, data):
            if TreeView.is_separator(model, item_iter, data):
                return
            item = model.get_value(item_iter, 0)
            markup = template is not None
            text = callout(item)
            if isinstance(template, str):
                text = template.format(self.clean(text, markup=True))
            else:
                text = self.clean(text)
            cell.set_property("markup", str(text))

        return internal

    def image_func(self, call, pixmaps=None):
        """Wrap, wrap wrap the func"""
        callout = self.get_callout(call)

        def internal(column, cell, model, item_iter, data):
            if TreeView.is_separator(model, item_iter, data):
                return
            item = model.get_value(item_iter, 0)
            icon = callout(item)
            # The or blank asks for the default icon from the pixmaps
            if isinstance(icon or "", str) and pixmaps:
                # Expect a Gnome theme icon
                icon = pixmaps.get(icon)
            elif icon:
                icon = pixmaps.apply_filters(icon)

            cell.set_property("pixbuf", icon)
            cell.set_property("visible", True)

        return internal

    def default_text(self, item):
        """Default text return for markup."""
        return default(item, "name", str(item))

    def default_icon(self, item):
        """Default icon return, pixbuf or gnome theme name"""
        return default(item, "icon", None)


class Separator:
    """Representation of a separator in a list"""



"""
inkvn (extension-curve) (2025/1/21)

description: Linearity Curve / Vectornator file importer for Inkscape

! what DOESN'T work (2025/03/15): textOnPath, grid, brush stroke, marker(arrow)
! and other features I missed
"""

import os
import sys

import inkex

from .reader.read import CurveReader
from .svg.convert import CurveConverter


HERE = os.path.dirname(__file__) or "."
# This is suggested by https://docs.python-guide.org/writing/structure/.
sys.path.insert(0, os.path.abspath(os.path.join(HERE, "..")))


class CurveInput(inkex.InputExtension):
    """Open and convert .curve / .vectornator files."""

    # copied from inkaf
    def load(self, stream):
        converter = CurveConverter()
        converter.convert(CurveReader(stream))
        return converter.doc


def main():
    CurveInput().run()



"""
VI extractors

reads Vectornator / Linearity Curve JSON data from .vectornator or .curve files.
"""

import base64
import io
import json
import logging
import zipfile
from typing import Any, Dict


def read_json_from_zip(
    archive: zipfile.ZipFile, file_name: str, max_depth: int = 2
) -> Dict[str, Any]:
    """Reads JSON file from zip, handling nested folders and embedded zip files."""
    try:
        # Get archive name
        archive_name = archive.filename

        # Check if the file exists at the top level
        if file_name in archive.namelist():
            with archive.open(file_name) as f:
                return json.load(f)

        # Search for the file within nested folders if not found at the top level
        for name in archive.namelist():
            if name.endswith(file_name):
                with archive.open(name) as f:
                    return json.load(f)

            # extract .curve / .vectornator if it's there (.vectornator in .vectornator)
            if name.endswith((".curve", ".vectornator")) and max_depth > 0:
                with archive.open(name) as nested_zip_file:
                    with zipfile.ZipFile(
                        io.BytesIO(nested_zip_file.read()), "r"
                    ) as nested_zip:
                        return read_json_from_zip(nested_zip, file_name, max_depth - 1)

        # Raise an error if the file is not found anywhere in the archive
        raise FileNotFoundError(
            f"File '{file_name}' not found in the zip archive '{archive_name}'."
        )

    except (json.JSONDecodeError, FileNotFoundError) as e:
        logging.error(
            f"Archive name: {archive_name}, Failed to read or parse JSON file '{file_name}': {e}"
        )
        raise  # Re-raise the exception after logging.


def read_dat_from_zip(
    archive: zipfile.ZipFile, file_name: str, max_depth: int = 2
) -> str:
    """Encode dat (bitmap) file from zip (Vectornator file) in Base64 string."""
    try:
        # Get archive name
        archive_name = archive.filename

        if file_name in archive.namelist():
            with archive.open(file_name) as f:
                return base64.b64encode(f.read()).decode("utf-8")

        for name in archive.namelist():
            if name.endswith(file_name):
                with archive.open(name) as f:
                    return base64.b64encode(f.read()).decode("utf-8")

            # extract .curve / .vectornator if it's there
            if name.endswith((".curve", ".vectornator")) and max_depth > 0:
                with archive.open(name) as nested_zip_file:
                    with zipfile.ZipFile(
                        io.BytesIO(nested_zip_file.read()), "r"
                    ) as nested_zip:
                        return read_dat_from_zip(nested_zip, file_name, max_depth - 1)

        raise FileNotFoundError(
            f"File '{file_name}' not found in the zip archive '{archive_name}'."
        )

    except Exception as e:
        logging.error(
            f"Archive name: {archive_name}, Failed to read or encode bitmap file '{file_name}': {e}"
        )
        raise


def extract_manifest(archive: zipfile.ZipFile) -> Dict[str, Any]:
    """Extract and parse the Manifest.json."""
    return read_json_from_zip(archive, "Manifest.json")


def extract_document(
    archive: zipfile.ZipFile, manifest: Dict[str, Any]
) -> Dict[str, Any]:
    """Extract and parse the Document.json specified in the Manifest."""
    document_name = manifest.get("documentJSONFilename", "Document.json")
    return read_json_from_zip(archive, document_name)


def extract_drawing_data(document: Dict[str, Any]) -> Dict[str, Any]:
    """Return the 'drawing' data from the Document.json."""
    return document.get("drawing")


def extract_gid_json(archive: zipfile.ZipFile, artboard_path: str) -> Dict[str, Any]:
    """Extract and parse a GUID JSON file (artboard)."""
    return read_json_from_zip(archive, artboard_path)



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2011 Nikita Kitaev
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
"""
Simplifies SVG files in preparation for sif export.
"""

from inkex.localization import inkex_gettext as _
from inkex.base import TempDirMixin

import inkex
from inkex import (
    Group,
    PathElement,
    ShapeElement,
    Anchor,
    Switch,
    SvgDocumentElement,
    Transform,
)

###### Utility Classes ####################################


class MalformedSVGError(Exception):
    """Raised when the SVG document is invalid or contains unsupported features"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return _(
            """SVG document is invalid or contains unsupported features

Error message: %s

The SVG to Synfig converter is designed to handle SVG files that were created using Inkscape. Unsupported features are most likely to occur in SVG files written by other programs.
"""
        ) % repr(self.value)


###### Utility Functions ##################################

### Path related


def fuse_subpaths(path: inkex.Path):
    """Fuse subpaths of a path. Should only be used on unstroked paths.
    The idea is to replace every moveto by a lineto, and then walk all the extra lines
    backwards to get the same fill. For unfilled paths, this gives visually good results
    in cases with not-too-complex paths, i.e. no intersections.
    There may be extra zero-length Lineto commands."""

    result = inkex.Path()
    return_stack = []
    for i, seg in enumerate(path.proxy_iterator()):
        if seg.letter not in "zZmM":
            result.append(seg.command)
        elif i == 0:
            result.append(seg.command)
            first = seg.end_point
        else:
            # Add a line instead of the ZoneClose / Move command, and store
            # the initial position.
            return_to = seg.previous_end_point
            if seg.letter in "mM":
                if seg.previous_end_point != first:  # only close if needed
                    # all paths must be closed for this algorithm to work. Since
                    # the path doesn't have a stroke, it's visually irrellevant.
                    result.append(inkex.paths.Line(*first))
                    return_to = first

                return_stack += [return_to]
            result.append(inkex.paths.Line(*seg.end_point))

            first = seg.end_point

    # also close the last subpath
    return_stack += [first]

    # now apply the return stack backwards
    for point in return_stack[::-1]:
        result.append(inkex.paths.Line(*point))

    return result


def split_fill_and_stroke(path_node):
    """Split a path into two paths, one filled and one stroked

    Returns a the list [fill, stroke], where each is the XML element of the
    fill or stroke, or None.
    """
    style = dict(inkex.Style.parse_str(path_node.get("style", "")))

    # If there is only stroke or only fill, don't split anything
    if "fill" in style and style["fill"] == "none":
        if "stroke" not in style or style["stroke"] == "none":
            return [None, None]  # Path has neither stroke nor fill
        else:
            return [None, path_node]
    if "stroke" not in style.keys() or style["stroke"] == "none":
        return [path_node, None]

    group = Group()
    fill = group.add(PathElement())
    stroke = group.add(PathElement())

    d = path_node.pop("d")
    if d is None:
        raise AssertionError("Cannot split stroke and fill of non-path element")

    nodetypes = path_node.pop("sodipodi:nodetypes", None)
    path_id = path_node.pop("id", str(id(path_node)))
    transform = path_node.pop("transform", None)
    path_node.pop("style")

    # Pass along all remaining attributes to the group
    for attrib_name, attrib_value in path_node.attrib.items():
        group.set(attrib_name, attrib_value)

    group.set("id", path_id)

    # Next split apart the style attribute
    style_group = {}
    style_fill = {"stroke": "none", "fill": "#000000"}
    style_stroke = {"fill": "none", "stroke": "none"}

    for key in style.keys():
        if key.startswith("fill"):
            style_fill[key] = style[key]
        elif key.startswith("stroke"):
            style_stroke[key] = style[key]
        elif key.startswith("marker"):
            style_stroke[key] = style[key]
        elif key.startswith("filter"):
            style_group[key] = style[key]
        else:
            style_fill[key] = style[key]
            style_stroke[key] = style[key]

    if len(style_group) != 0:
        group.set("style", str(inkex.Style(style_group)))

    fill.set("style", str(inkex.Style(style_fill)))
    stroke.set("style", str(inkex.Style(style_stroke)))

    # Finalize the two paths
    fill.set("d", d)
    stroke.set("d", d)
    if nodetypes is not None:
        fill.set("sodipodi:nodetypes", nodetypes)
        stroke.set("sodipodi:nodetypes", nodetypes)
    fill.set("id", path_id + "-fill")
    stroke.set("id", path_id + "-stroke")
    if transform is not None:
        fill.set("transform", transform)
        stroke.set("transform", transform)

    # Replace the original node with the group
    path_node.getparent().replace(path_node, group)

    return [fill, stroke]


### Object related


def propagate_attribs(
    node, parent_style={}, parent_transform=[[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]]
):
    """Propagate style and transform to remove inheritance"""

    # Don't enter non-graphical portions of the document
    if not isinstance(node, (ShapeElement, SvgDocumentElement)):
        return

    # Compose the transformations
    if isinstance(node, SvgDocumentElement) and node.get("viewBox"):
        vx, vy, vw, vh = [get_dimension(x) for x in node.get_viewbox()]
        dw = get_dimension(node.get("width", vw))
        dh = get_dimension(node.get("height", vh))
        this_transform = Transform(translate=(-vx, -vy), scale=(dw / vw, dh / vh))
        del node.attrib["viewBox"]
    else:
        this_transform = Transform(parent_transform)

    this_transform @= node.transform

    # Compose the style attribs
    this_style = dict(inkex.Style.parse_str(node.get("style", "")))
    remaining_style = {}  # Style attributes that are not propagated

    non_propagated = ["filter"]  # Filters should remain on the topmost ancestor
    for key in non_propagated:
        if key in this_style.keys():
            remaining_style[key] = this_style[key]
            del this_style[key]

    # Create a copy of the parent style, and merge this style into it
    parent_style_copy = parent_style.copy()
    parent_style_copy.update(this_style)
    this_style = parent_style_copy

    # Merge in any attributes outside of the style
    style_attribs = ["fill", "stroke"]
    for attrib in style_attribs:
        if node.get(attrib):
            this_style[attrib] = node.get(attrib)
            del node.attrib[attrib]

    if isinstance(node, (SvgDocumentElement, Group, Anchor, Switch)):
        # Leave only non-propagating style attributes
        if remaining_style:
            node.style = remaining_style
        else:
            if "style" in node.keys():
                del node.attrib["style"]

        # Remove the transform attribute
        if "transform" in node.keys():
            del node.attrib["transform"]

        # Continue propagating on subelements
        for child in node.iterchildren():
            propagate_attribs(child, this_style, this_transform)
    else:
        # This element is not a container

        # Merge remaining_style into this_style
        this_style.update(remaining_style)

        # Set the element's style and transform attribs
        node.style = this_style
        node.transform = this_transform


### Style related


def get_dimension(s="1024"):
    """Convert an SVG length string from arbitrary units to pixels"""
    return inkex.units.convert_unit(s, "px")



#!/usr/bin/env python3

# Copyright (C) 2024 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Object types for the binary parser
"""

from __future__ import annotations

from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Union


@dataclass
class Field:
    type: int
    value: Any = None
    parent: Optional[AFObject] = None


@dataclass
class AFObjectMetadata:
    id: int
    tag: str
    status: MetadataStatus


class ObjectStatus(Enum):
    NULL = 0
    SHARED = 1
    LINK = 2
    NON_SHARED = 3
    ROOT = 4


class MetadataStatus(Enum):
    TAG = 0
    TAG_ID = 1


@dataclass
class AFObject:
    id: int
    types: List[AFObjectMetadata]
    status: ObjectStatus

    def get_type(self) -> Optional[str]:
        return self.types[0].tag if self.types else None

    def append(self, tag: Optional[str], value: Any) -> None: ...

    def get(self, key: str | int, default: Any = None) -> Any:
        try:
            return self[key]
        except KeyError:
            return default

    @abstractmethod
    def __getitem__(self, key: Any) -> Any: ...

    @abstractmethod
    def __contains__(self, key: Any) -> bool: ...

    @abstractmethod
    def __len__(self) -> int: ...


@dataclass
class AFDictObject(AFObject):
    fields: Dict[str, Union[Field, List[Field]]]

    def append(self, tag: Optional[str], value: Any) -> None:
        assert tag is not None, "AFDictObject entries need tags"
        if tag in self.fields:
            fld = self.fields[tag]
            if isinstance(fld, list):
                fld.append(value)
            else:
                self.fields[tag] = [fld, value]
        else:
            self.fields[tag] = value

    def __getitem__(self, key: str) -> Any:
        field = self.fields[key]
        return field[-1].value if isinstance(field, list) else field.value

    def __contains__(self, key: str) -> bool:
        return key in self.fields

    def __len__(self) -> int:
        return len(self.fields)


@dataclass
class AFListObject(AFObject):
    fields: List[Field]

    def append(self, tag: Optional[str], value: Any) -> None:
        self.fields.append(value)

    def __getitem__(self, index: int) -> Any:
        if index not in self:
            raise KeyError(index)
        field = self.fields[index]
        return field[-1].value if isinstance(field, list) else field.value

    def __contains__(self, index: int) -> bool:
        return 0 <= index < len(self.fields)

    def __len__(self) -> int:
        return len(self.fields)


@dataclass
class EnumT:  # TODO: What is this?
    id: int
    version: int


@dataclass
class Curve12:
    doubles: float
    unsigned_ints: Tuple[int, int, int, int]


@dataclass
class Curve16:
    doubles: float
    unsigned_ints: Tuple[int, int]


class Curve18Type(Enum):
    CONTROL_POINT = 0
    SHARP = 1
    SMOOTH = 2
    SMART = 3
    SMOOTH_SHARP = 4  # TODO: Sharp with control points?


@dataclass
class Curve18:
    point: Tuple[float, float]
    type: Curve18Type
    index: int


@dataclass
class UnknownStruct:
    data: bytes


@dataclass
class FlagsT:
    version: int
    flags: bytes


@dataclass
class EmbeddedData:
    tag: str
    size: int
    data: str


@dataclass
class Document(AFDictObject):
    doc_format: Dict[str, Any]



#!/usr/bin/env python3
"""
Convert PNG to Tiff using Raster Output extension.
"""

import io
import inkex


class TiffOutput(inkex.RasterOutputExtension):
    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument("--compression", default=None)
        pars.add_argument("--quality", type=int, default=90)

    def save(self, stream):
        tempstream = io.BytesIO()
        self.img.save(
            tempstream,
            format="tiff",
            compression=(self.options.compression or None),
            dpi=self.img.info["dpi"],
        )
        stream.write(tempstream.getvalue())
        # TODO: Add other fields such as copyright etc.


if __name__ == "__main__":
    TiffOutput().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2008 Jonas Termeau - jonas.termeau **AT** gmail.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Thanks to:
#
# Bernard Gray - bernard.gray **AT** gmail.com (python helping)
# Jamie Heames (english translation issues)
# ~suv (bug report in v2.3)
# http://www.gutenberg.eu.org/publications/ (9x9 margins settings)
#
"""
This basic extension allows you to automatically draw guides in inkscape.
"""

from math import cos, sin, sqrt
import re

import inkex

from inkex.localization import inkex_gettext as _


class GuidesOpts:
    """Manager of current-page-related values for GuidesCreator"""

    # pylint: disable=too-few-public-methods
    def __init__(self, svg: inkex.SvgDocumentElement) -> None:
        # get page bounds
        self.pages = svg.namedview.get_pages()
        self.viewbox = svg.get_viewbox()
        # in case the viewbox attribute is not set, fall back to viewport
        self.viewbox[2:4] = svg.viewbox_width, svg.viewbox_height
        self.set_page(1)

    def set_page(self, pagenumber):
        """Update guide origin and width/height based on page number (1-indexed)"""
        self.pagenumber = pagenumber
        pagenumber = pagenumber - 1
        if 0 <= pagenumber < len(self.pages):
            self.page_origin = (self.pages[pagenumber].x, self.pages[pagenumber].y)
            self.width = self.pages[pagenumber].width
            self.height = self.pages[pagenumber].height
        else:
            raise ValueError("Invalid page number")

        # getting edges coordinates
        self.orientation = ((round(self.height, 4), 0), (0, round(self.width, 4)))


class GuidesCreator(inkex.EffectExtension):
    """Create a set of guides based on the given options"""

    def add_arguments(self, pars):
        pars.add_argument(
            "--pages",
            type=self.arg_number_ranges(),
            help='On which pages the guides are created, e.g. "1, 2, 4-6, 8-". '
            "Default: All pages.",
            default="1-",
        )
        pars.add_argument(
            "--tab",
            type=self.arg_method("generate"),
            default="regular_guides",
            help="Type of guides to create.",
        )
        pars.add_argument("--guides_preset", default="custom", help="Preset")
        pars.add_argument(
            "--vertical_guides", type=int, default=2, help="Vertical guides"
        )
        pars.add_argument(
            "--horizontal_guides", type=int, default=3, help="Horizontal guides"
        )
        pars.add_argument(
            "--start_from_edges", type=inkex.Boolean, help="Start from edges"
        )
        pars.add_argument(
            "--ul", type=inkex.Boolean, default=False, help="Upper left corner"
        )
        pars.add_argument(
            "--ur", type=inkex.Boolean, default=False, help="Upper right corner"
        )
        pars.add_argument(
            "--ll", type=inkex.Boolean, default=False, help="Lower left corner"
        )
        pars.add_argument(
            "--lr", type=inkex.Boolean, default=False, help="Lower right corner"
        )
        pars.add_argument(
            "--margins_preset",
            default="custom",
            choices=[
                "custom",
                "book_left",
                "book_right",
                "book_alternating_right",
                "book_alternating_left",
            ],
            help="Margins preset",
        )
        pars.add_argument("--vert", type=int, default=0, help="Vert subdivisions")
        pars.add_argument("--horz", type=int, default=0, help="Horz subdivisions")
        pars.add_argument("--header_margin", type=int, default=10, help="Header margin")
        pars.add_argument("--footer_margin", type=int, default=10, help="Footer margin")
        pars.add_argument("--left_margin", type=int, default=10, help="Left margin")
        pars.add_argument("--right_margin", type=int, default=10, help="Right margin")
        pars.add_argument("--delete", type=inkex.Boolean, help="Delete existing guides")
        pars.add_argument(
            "--nodup", type=inkex.Boolean, help="Omit duplicated guides", default=True
        )

    def __init__(self):
        super().__init__()
        self.store: GuidesOpts = None

    def effect(self):
        if self.options.delete:
            for guide in self.svg.namedview.get_guides():
                guide.delete()

        self.store = GuidesOpts(self.svg)
        for i in self.options.pages(max(len(self.svg.namedview.get_pages()), 1)):
            self.store.set_page(i)
            self.options.tab()

    def generate_regular_guides(self):
        """Generate a regular set of guides"""
        preset = self.options.guides_preset
        from_edges = self.options.start_from_edges
        if preset == "custom":
            h_division = self.options.horizontal_guides
            v_division = self.options.vertical_guides
            if from_edges:
                v_division = v_division or 1
                h_division = h_division or 1

            self.draw_guides(v_division, from_edges, vert=True)
            self.draw_guides(h_division, from_edges, vert=False)

        elif preset == "golden":
            gold = (1 + sqrt(5)) / 2

            for fraction, index in zip([1 / gold, 1 - 1 / gold] * 2, [1, 1, 0, 0]):
                position = fraction * (self.store.width, self.store.height)[index]
                self.draw_guide(
                    (0, position) if index == 1 else (position, 0),
                    self.store.orientation[index],
                )

            if from_edges:
                self.draw_guides(1, True, vert=False)
                self.draw_guides(1, True, vert=True)

        elif ";" in preset:
            v_division = int(preset.split(";")[0])
            h_division = int(preset.split(";")[1])
            self.draw_guides(v_division, from_edges, vert=True)
            self.draw_guides(h_division, from_edges, vert=False)
        else:
            raise inkex.AbortExtension(_("Unknown guide preset: {}").format(preset))

    def generate_diagonal_guides(self):
        """Generate diagonal guides"""
        # Dimentions
        left, bottom = (0, 0)
        right, top = (self.store.width, self.store.height)

        # Diagonal angle
        angle = 45

        corner_guides = {
            "ul": ((left, top), (cos(angle), cos(angle))),
            "ur": ((right, top), (-sin(angle), sin(angle))),
            "ll": ((left, bottom), (-cos(angle), cos(angle))),
            "lr": ((right, bottom), (-sin(angle), -sin(angle))),
        }

        for key, (position, orientation) in corner_guides.items():
            if getattr(self.options, key):
                self.draw_guide(position, orientation)

    def generate_margins(self):
        """Generate margin guides"""

        if self.options.start_from_edges:
            # horizontal borders
            self.draw_guide((0, self.store.height), self.store.orientation[1])
            self.draw_guide((self.store.width, 0), self.store.orientation[1])

            # vertical borders
            self.draw_guide((0, self.store.height), self.store.orientation[0])
            self.draw_guide((self.store.width, 0), self.store.orientation[0])

        if self.options.margins_preset == "custom":
            margins = [
                (i / j if int(j) != 0 else None)
                for i, j in zip(
                    (
                        self.store.height * (self.options.header_margin - 1),  # header
                        self.store.height,  # footer
                        self.store.width,  # left
                        self.store.width * (self.options.right_margin - 1),  # right
                    ),
                    (
                        self.options.header_margin,
                        self.options.footer_margin,
                        self.options.left_margin,
                        self.options.right_margin,
                    ),
                )
            ]

        book_options = {
            "book_left": (8 / 9, 2 / 9, 2 / 9, 8 / 9),
            "book_right": (8 / 9, 2 / 9, 1 / 9, 7 / 9),
        }
        margins_preset = self.options.margins_preset
        if margins_preset.startswith("book_alternating"):
            margins_preset = (
                "book_left"
                if self.store.pagenumber % 2 == (1 if "left" in margins_preset else 0)
                else "book_right"
            )

        if margins_preset in book_options:
            margins = [
                i * j
                for i, j in zip(
                    book_options[margins_preset],
                    2 * [self.store.height] + 2 * [self.store.width],
                )
            ]

        y_header, y_footer, x_left, x_right = [
            i or j for i, j in zip(margins, [self.store.height, 0, 0, self.store.width])
        ]

        for length, position in zip(margins, [1, 1, 0, 0]):
            if length is None:
                continue
            self.draw_guide(
                (length, 0) if position == 0 else (0, length),
                self.store.orientation[position],
            )

        # setting up properties of the rectangle created between guides
        rectangle_height = y_header - y_footer
        rectangle_width = x_right - x_left

        for subdiv, vert, begin_from in zip(
            (self.options.horz, self.options.vert), (False, True), (y_footer, x_left)
        ):
            if subdiv != 0:
                self._draw_guides(
                    (rectangle_width, rectangle_height),
                    subdiv,
                    edges=0,
                    shift=begin_from,
                    vert=vert,
                )

    def draw_guides(self, division, edges, vert=False):
        """Draw a vertical or horizontal lines"""
        return self._draw_guides(
            (self.store.width, self.store.height), division, edges, vert=vert
        )

    def _draw_guides(self, vector, division, edges, shift=0, vert=False):
        if division <= 0:
            return

        # Vert controls both ort template and vector calculation
        def ort(x):
            return (x, 0) if vert else (0, x)

        var = int(bool(edges))
        for x in range(0, division - 1 + 2 * var):
            div = vector[not bool(vert)] / division
            position = round(div + (x - var) * div + shift, 4)
            orientation = round(vector[bool(vert)], 4)
            self.draw_guide(ort(position), ort(orientation))

    def draw_guide(self, position, orientation):
        """Draw the guides"""
        newpos = [
            position[0] + self.store.page_origin[0],
            -position[1] + self.store.height + self.store.page_origin[1],
        ]
        # orientations are computed in the pre-1.0 coordinate system
        orientation = [orientation[0], -orientation[1]]
        if self.options.nodup:
            self.svg.namedview.add_unique_guide(newpos, orientation)
        else:
            self.svg.namedview.add_guide(newpos, orientation)


if __name__ == "__main__":
    GuidesCreator().run()



#!/usr/bin/env python3
#
# Copyright (C) 2006 Jean-Francois Barraud, barraud@math.univ-lille1.fr
#               2021 Jonathan Neuhauser, jonathan.neuhauser@outlook.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# barraud@math.univ-lille1.fr
#
"""
This script deforms an object (the pattern) along other paths (skeletons)...
The first selected object is the pattern
the last selected ones are the skeletons.

Imagine a straight horizontal line L in the middle of the bounding box of the pattern.
Consider the normal bundle of L: the collection of all the vertical lines meeting L.
Consider this as the initial state of the plane; in particular, think of the pattern
as painted on these lines.

Now move and bend L to make it fit a skeleton, and see what happens to the normals:
they move and rotate, deforming the pattern.
"""

import copy
import math

import inkex
from inkex.bezier import tpoint
from inkex.paths import CubicSuperPath
from inkex.localization import inkex_gettext as _

import pathmodifier


class PatternAlongPath(pathmodifier.PathModifier):
    """Deform a path along a second path"""

    def add_arguments(self, pars):
        pars.add_argument(
            "-n", "--noffset", type=float, default=0.0, help="normal offset"
        )
        pars.add_argument(
            "-t", "--toffset", type=float, default=0.0, help="tangential offset"
        )
        pars.add_argument("-k", "--kind", type=str, default="")
        pars.add_argument(
            "-c",
            "--copymode",
            default="Single",
            help="repeat the path to fit deformer's length",
        )
        pars.add_argument("-p", "--space", type=float, default=0.0)
        pars.add_argument(
            "-v",
            "--vertical",
            type=inkex.Boolean,
            default=False,
            help="reference path is vertical",
        )
        pars.add_argument(
            "-d",
            "--duplicate",
            type=inkex.Boolean,
            default=True,
            help="duplicate pattern before deformation",
        )
        pars.add_argument("--tab", help="The selected UI-tab when OK was pressed")

    def apply_diffeomorphism(self, bpt, skelcomp, lengths, isclosed, vects=()):
        """
        The kernel of this stuff:
        bpt is a base point and for v in vectors, v'=v-p is a tangent vector at bpt.
        """
        s = bpt[0] - skelcomp[0][0]
        i, t = self.lengthtotime(s, lengths, isclosed)
        if i == len(skelcomp) - 1:
            x, y = tpoint(skelcomp[i - 1], skelcomp[i], 1 + t)
            dx = (skelcomp[i][0] - skelcomp[i - 1][0]) / lengths[-1]
            dy = (skelcomp[i][1] - skelcomp[i - 1][1]) / lengths[-1]
        else:
            x, y = tpoint(skelcomp[i], skelcomp[i + 1], t)
            dx = (skelcomp[i + 1][0] - skelcomp[i][0]) / lengths[i]
            dy = (skelcomp[i + 1][1] - skelcomp[i][1]) / lengths[i]

        vx = 0
        vy = bpt[1] - skelcomp[0][1]
        if self.options.wave:
            bpt[0] = x + vx * dx
            bpt[1] = y + vy + vx * dy
        else:
            bpt[0] = x + vx * dx - vy * dy
            bpt[1] = y + vx * dy + vy * dx

        for v in vects:
            vx = v[0] - skelcomp[0][0] - s
            vy = v[1] - skelcomp[0][1]
            if self.options.wave:
                v[0] = x + vx * dx
                v[1] = y + vy + vx * dy
            else:
                v[0] = x + vx * dx - vy * dy
                v[1] = y + vx * dy + vy * dx

    def effect(self):
        if len(self.options.ids) < 2:
            raise inkex.AbortExtension(_("This extension requires two selected paths."))

        self.options.wave = self.options.kind == "Ribbon"
        if self.options.copymode == "Single":
            self.options.repeat = False
            self.options.stretch = False
        elif self.options.copymode == "Repeated":
            self.options.repeat = True
            self.options.stretch = False
        elif self.options.copymode == "Single, stretched":
            self.options.repeat = False
            self.options.stretch = True
        elif self.options.copymode == "Repeated, stretched":
            self.options.repeat = True
            self.options.stretch = True

        patterns, skels = self.get_patterns_and_skeletons(True, self.options.duplicate)
        bboxes = [pattern.bounding_box() for pattern in patterns.values()]
        if None in bboxes:  # for texts, we can't compute the bounding box
            raise inkex.AbortExtension(_("Please convert texts to path first"))
        bbox = sum(bboxes, None)

        if self.options.vertical:
            # flipxy(bbox)...
            bbox = inkex.BoundingBox(-bbox.y, -bbox.x)

        width = bbox.width
        delta_x = width + self.options.space
        if delta_x < 0.01:
            raise inkex.AbortExtension(
                _(
                    "The total length of the pattern is too small\n"
                    "Please choose a larger object or set 'Space between copies' > 0"
                )
            )
        for pattern in patterns.values():
            if isinstance(pattern, inkex.PathElement):
                pattern.apply_transform()
                pattern.path = self._do_transform(
                    skels, pattern.path.to_superpath(), delta_x, bbox
                )

    def _do_transform(self, skeletons, p0, dx, bbox):
        if self.options.vertical:
            self.flipxy(p0)
        newp = []
        for skelnode in skeletons.values():
            skelnode.apply_transform()
            cur_skeleton = skelnode.path.to_superpath()
            if self.options.vertical:
                self.flipxy(cur_skeleton)
            for comp in cur_skeleton:
                path = copy.deepcopy(p0)
                skelcomp, lengths = self.linearize(comp)

                skel_closed = all(
                    [math.isclose(i, j) for i, j in zip(skelcomp[0], skelcomp[-1])]
                )

                length = sum(lengths)
                xoffset = skelcomp[0][0] - bbox.x.minimum + self.options.toffset
                yoffset = skelcomp[0][1] - bbox.y.center - self.options.noffset
                if self.options.repeat:
                    nb_copies = max(1, int(round((length + self.options.space) / dx)))
                    width = dx * nb_copies
                    if not skel_closed:
                        width -= self.options.space
                    bbox.x.maximum = bbox.x.minimum + width
                    new = []
                    for sub in path:
                        for _ in range(nb_copies):
                            new.append(copy.deepcopy(sub))
                            self.offset(sub, dx, 0)
                    path = new

                for sub in path:
                    self.offset(sub, xoffset, yoffset)

                if self.options.stretch:
                    if not bbox.width:
                        raise inkex.AbortExtension(
                            _(
                                "The 'stretch' option requires that the pattern must "
                                "have non-zero width :\nPlease edit the pattern width."
                            )
                        )
                    for sub in path:
                        self.stretch(sub, length / bbox.width, 1, skelcomp[0])

                for sub in path:
                    for ctlpt in sub:
                        self.apply_diffeomorphism(
                            ctlpt[1],
                            skelcomp,
                            lengths,
                            skel_closed,
                            (ctlpt[0], ctlpt[2]),
                        )

                if self.options.vertical:
                    self.flipxy(path)
                newp += path
        return CubicSuperPath(newp)


if __name__ == "__main__":
    PatternAlongPath().run()



#!/usr/bin/env python3
"""Rotate the colors in the selected elements"""

import inkex


class RgbBarrel(inkex.ColorExtension):
    """
    Cycle colors RGB -> BRG

    aka  Do a Barrel Roll!
    """

    def modify_color(self, name, color):
        return inkex.Color((color.blue, color.red, color.green))


if __name__ == "__main__":
    RgbBarrel().run()
